import click

find = click.option(
    '--find-query-port',
    default=False,
    is_flag=True,
    help='(Attempt to) find the query port for each server'
)
gamedig_bin = click.option(
    '--gamedig-bin',
    type=str,
    default='/usr/bin/gamedig',
    help='Path to gamedig binary'
)
gamedig_concurrency = click.option(
    '--gamedig-concurrency',
    type=int,
    default=12,
    help='Number of gamedig queries to run in parallel'
)



import logging
import sys

import click

from GameserverLister.commands.options import common
from GameserverLister.common.logger import logger
from GameserverLister.common.types import ValveGame, ValvePrincipal
from GameserverLister.games.valve import VALVE_GAME_CONFIGS
from GameserverLister.listers import ValveServerLister


@click.command
@click.option(
    '-g',
    '--game',
    type=click.Choice(ValveGame),
    required=True,
    help='Game to list servers for'
)
@click.option(
    '-p',
    '--principal',
    type=click.Choice(ValvePrincipal),
    default=ValvePrincipal.VALVE,
    help='Principal server to query'
)
@click.option(
    '-f',
    '--filter',
    'filters',
    type=str,
    default='',
    help='Filter to apply to server list'
)
@click.option(
    '-t',
    '--timeout',
    type=int,
    default=5,
    help='Timeout to use for principal query'
)
@click.option(
    '-m',
    '--max-pages',
    type=int,
    default=10,
    help='Maximum number of pages to retrieve from the server list (per region)'
)
@common.expired_ttl
@common.list_dir
@common.recover
@common.add_links
@common.debug
def run(
        game: ValveGame,
        principal: str,
        filters: str,
        timeout: int,
        max_pages: int,
        expired_ttl: int,
        recover: bool,
        add_links: bool,
        list_dir: str,
        debug: bool
):
    logging.basicConfig(level=logging.DEBUG if debug else logging.INFO, stream=sys.stdout,
                        format='%(asctime)s %(levelname)-8s %(message)s')
    logger.info(f'Listing servers for {game} via valve/{principal}')

    # Set principal
    available_principals = VALVE_GAME_CONFIGS[game].principals
    if principal not in VALVE_GAME_CONFIGS[game].principals:
        # Given principal is invalid => use default principal
        principal = available_principals[0]

    lister = ValveServerLister(
        game,
        principal,
        timeout,
        filters,
        max_pages,
        expired_ttl,
        recover,
        add_links,
        list_dir
    )

    before = len(lister.servers)
    lister.update_server_list()
    removed, recovered = lister.remove_expired_servers()
    lister.write_to_file()

    logger.info(f'Server list updated ('
                f'total: {len(lister.servers)}, '
                f'added: {len(lister.servers) + removed - before}, '
                f'removed: {removed}, '
                f'recovered: {recovered})')



from typing import Dict

from GameserverLister.common.types import BattlelogGame

BATTLELOG_GAME_BASE_URIS: Dict[BattlelogGame, str] = {
    BattlelogGame.BF3: 'https://battlelog.battlefield.com/bf3/servers/getAutoBrowseServers/',
    BattlelogGame.BF4: 'https://battlelog.battlefield.com/bf4/servers/getServers/pc/',
    BattlelogGame.BFH: 'https://battlelog.battlefield.com/bfh/servers/getServers/pc/',
    BattlelogGame.MOHWF: 'https://battlelog.battlefield.com/mohw/servers/getAutoBrowseServers/'
}



import click

from GameserverLister.commands import gamespy, battlelog, unreal2, valve, quake3, gametools, bfbc2, medalofhonor


@click.group()
def cli():
    pass


cli.add_command(battlelog.run, 'battlelog')
cli.add_command(bfbc2.run, 'bfbc2')
cli.add_command(gamespy.run, 'gamespy')
cli.add_command(gametools.run, 'gametools')
cli.add_command(medalofhonor.run, 'medalofhonor')
cli.add_command(quake3.run, 'quake3')
cli.add_command(unreal2.run, 'unreal2')
cli.add_command(valve.run, 'valve')

if __name__ == '__main__':
    cli()



import json
from datetime import datetime, timedelta
from typing import Union, Optional, Any, List

from GameserverLister.common.constants import UNIX_EPOCH_START
from GameserverLister.common.weblinks import WebLink


class ObjectJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        return obj.dump()


class Server:
    uid: str
    # Only optional because lists may still contain entries without this attribute
    first_seen_at: Optional[datetime]
    last_seen_at: datetime
    links: List[WebLink]

    def __init__(
            self,
            guid: str,
            links: Union[List[WebLink], WebLink],
            first_seen_at: Optional[datetime] = datetime.now().astimezone(),
            last_seen_at: datetime = datetime.now().astimezone(),
    ):
        self.uid = guid
        self.first_seen_at = first_seen_at
        self.last_seen_at = last_seen_at
        self.links = links if isinstance(links, list) else [links]

    def add_links(self, links: Union[List[WebLink], WebLink]) -> None:
        link_list = links if isinstance(links, list) else [links]
        for web_link in link_list:
            web_sites_self = [web_link.site for web_link in self.links]
            if web_link.site not in web_sites_self:
                self.links.append(web_link)
            else:
                index = web_sites_self.index(web_link.site)
                self.links[index].update(web_link)

    def trim(self, expired_ttl: float) -> None:
        """
        "Trim"/remove expired attributes (no longer valid, old via status)
        :param expired_ttl: Number of hours attributes remain valid after being last updated
        :return:
        """
        self.links = [link for link in self.links if not link.is_expired(expired_ttl)]

    def update(self, updated: 'Server') -> None:
        self.last_seen_at = updated.last_seen_at
        # Merge links "manually"
        self.add_links(updated.links)

    @staticmethod
    def load(parsed: dict) -> Union['Server', dict]:
        pass

    @staticmethod
    def is_json_repr(parsed: dict) -> bool:
        pass

    # Should be called "__dict__" but that confused the PyCharm debugger and
    # makes it impossible to inspect any instance variables
    # https://youtrack.jetbrains.com/issue/PY-43955
    def dump(self) -> dict:
        pass

    def __iter__(self):
        yield from self.dump().items()

    def __str__(self):
        return json.dumps(dict(self), cls=ObjectJSONEncoder)

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, type(self)) and \
               other.uid == self.uid and \
               other.first_seen_at == self.first_seen_at and \
               other.last_seen_at == self.last_seen_at


class QueryableServer(Server):
    ip: str
    query_port: int

    def __init__(
            self,
            guid: str,
            ip: str,
            query_port: int,
            links: Union[List[WebLink], WebLink],
            first_seen_at: Optional[datetime] = datetime.now().astimezone(),
            last_seen_at: datetime = datetime.now().astimezone()
    ):
        super().__init__(guid, links, first_seen_at, last_seen_at)
        self.ip = ip
        self.query_port = query_port

    def update(self, updated: 'QueryableServer') -> None:
        Server.update(self, updated)
        self.ip = updated.ip
        self.query_port = updated.query_port

    @staticmethod
    def is_json_repr(parsed: dict) -> bool:
        return 'guid' in parsed and 'ip' in parsed and 'queryPort' in parsed

    def __eq__(self, other: Any) -> bool:
        return Server.__eq__(self, other) and other.ip == self.ip and other.query_port == self.query_port


class ViaStatus:
    principal: str
    first_seen_at: datetime
    last_seen_at: datetime

    def __init__(self, principal: str, first_seen_at: datetime = datetime.now().astimezone(),
                 last_seen_at: datetime = datetime.now().astimezone()):
        self.principal = principal
        self.first_seen_at = first_seen_at
        self.last_seen_at = last_seen_at

    def is_expired(self, expired_ttl: float) -> bool:
        return datetime.now().astimezone() > self.last_seen_at + timedelta(hours=expired_ttl)

    def update(self, updated: 'ViaStatus') -> None:
        self.last_seen_at = updated.last_seen_at

    @staticmethod
    def load(parsed: dict) -> 'ViaStatus':
        first_seen_at = datetime.fromisoformat(parsed['firstSeenAt'])
        last_seen_at = datetime.fromisoformat(parsed['lastSeenAt'])

        return ViaStatus(
            parsed['principal'],
            first_seen_at,
            last_seen_at
        )

    @staticmethod
    def is_json_repr(parsed: dict) -> bool:
        return 'principal' in parsed and 'firstSeenAt' in parsed and 'lastSeenAt' in parsed

    def dump(self) -> dict:
        return {
            'principal': self.principal,
            'firstSeenAt': self.first_seen_at.isoformat(),
            'lastSeenAt': self.last_seen_at.isoformat()
        }

    def __eq__(self, other):
        return isinstance(other, ViaStatus) and \
               other.principal == self.principal and \
               other.first_seen_at == self.first_seen_at and \
               other.last_seen_at == self.last_seen_at

    def __iter__(self):
        yield from self.dump().items()

    def __str__(self):
        return json.dumps(dict(self))

    def __repr__(self):
        return self.__str__()


class ClassicServer(QueryableServer):
    """
    Server for "classic" games whose principals which return a server list
    containing ips and query ports of game servers (GameSpy, Quake3)
    """
    via: List[ViaStatus]

    def __init__(
            self,
            guid: str,
            ip: str,
            query_port: int,
            via: Union[List[ViaStatus], ViaStatus],
            first_seen_at: Optional[datetime] = datetime.now().astimezone(),
            last_seen_at: datetime = datetime.now().astimezone()
    ):
        # Leave link list empty for now (query port is usually not sufficient to build any links)
        super().__init__(guid, ip, query_port, [], first_seen_at, last_seen_at)
        self.via = via if isinstance(via, list) else [via]

    def trim(self, expired_ttl: float) -> None:
        super().trim(expired_ttl)
        self.via = [via for via in self.via if not via.is_expired(expired_ttl)]

    def update(self, updated: 'ClassicServer') -> None:
        QueryableServer.update(self, updated)
        # Merge via statuses "manually"
        for via_status in updated.via:
            via_principals_self = [via_status.principal for via_status in self.via]
            if via_status.principal not in via_principals_self:
                self.via.append(via_status)
            else:
                index = via_principals_self.index(via_status.principal)
                self.via[index].update(via_status)

    @staticmethod
    def load(parsed: dict) -> Union['ClassicServer', dict]:
        # Return data as is if it's not a JSON representation
        if not ClassicServer.is_json_repr(parsed):
            return parsed

        first_seen_at = datetime.fromisoformat(parsed['firstSeenAt']) \
            if parsed.get('firstSeenAt') is not None else None
        last_seen_at = datetime.fromisoformat(parsed['lastSeenAt']) \
            if parsed.get('lastSeenAt') is not None else UNIX_EPOCH_START
        via = [
            ViaStatus.load(via_parsed) for via_parsed in parsed.get('via', []) if
            ViaStatus.is_json_repr(via_parsed)
        ]

        server = ClassicServer(parsed['guid'], parsed['ip'], parsed['queryPort'], via, first_seen_at, last_seen_at)
        server.links = [
            WebLink.load(link_parsed) for link_parsed in parsed.get('links', [])
            if WebLink.is_json_repr(link_parsed)
        ]

        return server

    def dump(self) -> dict:
        return {
            'guid': self.uid,
            'ip': self.ip,
            'queryPort': self.query_port,
            'firstSeenAt': self.first_seen_at.isoformat() if self.first_seen_at is not None else self.first_seen_at,
            'lastSeenAt': self.last_seen_at.isoformat(),
            'via': [via_status.dump() for via_status in self.via],
            'links': [link.dump() for link in self.links]
        }

    def __eq__(self, other):
        return QueryableServer.__eq__(self, other) and all(via_status in self.via for via_status in other.via)


class FrostbiteServer(QueryableServer):
    """
    Server for Frostbite-era games whose server lists are centralized (contain all relevant server info rather than
    just the query port)
    """
    name: str
    game_port: int
    last_queried_at: Optional[datetime]

    def __init__(
            self,
            guid: str,
            name: str,
            ip: str,
            game_port: int,
            query_port: int = -1,
            first_seen_at: Optional[datetime] = datetime.now().astimezone(),
            last_seen_at: datetime = datetime.now().astimezone(),
            last_queried_at: Optional[datetime] = None
    ):
        # Leave link list empty for now (adding links is optional after all)
        super().__init__(guid, ip, query_port, [], first_seen_at, last_seen_at)
        self.name = name
        self.game_port = game_port
        self.last_queried_at = last_queried_at

    def update(self, updated: 'FrostbiteServer') -> None:
        Server.update(self, updated)
        # Cannot use parent's update function here, since that would always overwrite the query port
        # (which may be -1 for updated if query failed or was not attempted/enabled)
        self.ip = updated.ip
        if updated.query_port != -1:
            self.query_port = updated.query_port
        self.name = updated.name
        self.game_port = updated.game_port
        # Only update timestamp if not none (don't reset)
        if updated.last_queried_at is not None:
            self.last_queried_at = updated.last_queried_at

    @staticmethod
    def load(parsed: dict) -> Union['FrostbiteServer', dict]:
        if not FrostbiteServer.is_json_repr(parsed):
            return parsed

        first_seen_at = datetime.fromisoformat(parsed['firstSeenAt']) \
            if parsed.get('firstSeenAt') is not None else None
        last_seen_at = datetime.fromisoformat(parsed['lastSeenAt']) \
            if parsed.get('lastSeenAt') is not None else UNIX_EPOCH_START
        last_queried_at = datetime.fromisoformat(parsed['lastQueriedAt']) \
            if parsed.get('lastQueriedAt') not in [None, ''] else None

        server = FrostbiteServer(
            parsed['guid'],
            parsed['name'],
            parsed['ip'],
            parsed['gamePort'],
            parsed['queryPort'],
            first_seen_at,
            last_seen_at,
            last_queried_at
        )
        server.links = [
            WebLink.load(link_parsed) for link_parsed in parsed.get('links', [])
            if WebLink.is_json_repr(link_parsed)
        ]

        return server

    @staticmethod
    def is_json_repr(parsed: dict) -> bool:
        return QueryableServer.is_json_repr(parsed) and 'name' in parsed and 'gamePort' in parsed

    def dump(self) -> dict:
        return {
            'guid': self.uid,
            'name': self.name,
            'ip': self.ip,
            'gamePort': self.game_port,
            'queryPort': self.query_port,
            'firstSeenAt': self.first_seen_at.isoformat() if self.first_seen_at is not None else self.first_seen_at,
            'lastSeenAt': self.last_seen_at.isoformat(),
            'lastQueriedAt': self.last_queried_at.isoformat() if self.last_queried_at is not None else self.last_queried_at,
            'links': [link.dump() for link in self.links]
        }


class BadCompany2Server(FrostbiteServer):
    lid: int
    gid: int

    def __init__(
            self,
            guid: str,
            name: str,
            lid: int,
            gid: int,
            ip: str,
            game_port: int,
            query_port: int = -1,
            first_seen_at: Optional[datetime] = datetime.now().astimezone(),
            last_seen_at: datetime = datetime.now().astimezone(),
            last_queried_at: Optional[datetime] = None
    ):
        # Leave link list empty for now (adding links is optional after all)
        super().__init__(guid, name, ip, game_port, query_port, first_seen_at, last_seen_at, last_queried_at)
        self.lid = lid
        self.gid = gid

    @staticmethod
    def load(parsed: dict) -> Union['BadCompany2Server', dict]:
        # Will change to own validation in future
        if not FrostbiteServer.is_json_repr(parsed):
            return parsed

        first_seen_at = datetime.fromisoformat(parsed['firstSeenAt']) \
            if parsed.get('firstSeenAt') is not None else None
        last_seen_at = datetime.fromisoformat(parsed['lastSeenAt']) \
            if parsed.get('lastSeenAt') is not None else UNIX_EPOCH_START
        last_queried_at = datetime.fromisoformat(parsed['lastQueriedAt']) \
            if parsed.get('lastQueriedAt') not in [None, ''] else None

        server = BadCompany2Server(
            parsed['guid'],
            parsed['name'],
            parsed.get('lid', -1),
            parsed.get('gid', -1),
            parsed['ip'],
            parsed['gamePort'],
            parsed['queryPort'],
            first_seen_at,
            last_seen_at,
            last_queried_at
        )
        server.links = [
            WebLink.load(link_parsed) for link_parsed in parsed.get('links', [])
            if WebLink.is_json_repr(link_parsed)
        ]

        return server

    def dump(self) -> dict:
        return {
            'guid': self.uid,
            'name': self.name,
            'ip': self.ip,
            'gamePort': self.game_port,
            'queryPort': self.query_port,
            'lid': self.lid,
            'gid': self.gid,
            'firstSeenAt': self.first_seen_at.isoformat() if self.first_seen_at is not None else self.first_seen_at,
            'lastSeenAt': self.last_seen_at.isoformat(),
            'lastQueriedAt': self.last_queried_at.isoformat() if self.last_queried_at is not None else self.last_queried_at,
            'links': [link.dump() for link in self.links]
        }


class GametoolsServer(Server):
    name: str

    def __init__(
            self,
            game_id: str,
            name: str,
            first_seen_at: Optional[datetime] = datetime.now().astimezone(),
            last_seen_at: datetime = datetime.now().astimezone(),
    ):
        # Leave link list empty for now (adding links is optional after all)
        super().__init__(game_id, [], first_seen_at, last_seen_at)
        self.name = name

    def update(self, updated: 'GametoolsServer') -> None:
        Server.update(self, updated)
        self.name = updated.name

    @staticmethod
    def load(parsed: dict) -> Union['GametoolsServer', dict]:
        if not GametoolsServer.is_json_repr(parsed):
            return parsed

        first_seen_at = datetime.fromisoformat(parsed['firstSeenAt']) \
            if parsed.get('firstSeenAt') is not None else None
        last_seen_at = datetime.fromisoformat(parsed['lastSeenAt']) \
            if parsed.get('lastSeenAt') is not None else UNIX_EPOCH_START

        server = GametoolsServer(parsed['gameId'], parsed['name'], first_seen_at, last_seen_at)
        server.links = [
            WebLink.load(link_parsed) for link_parsed in parsed.get('links', [])
            if WebLink.is_json_repr(link_parsed)
        ]

        return server

    @staticmethod
    def is_json_repr(parsed: dict) -> bool:
        return 'gameId' in parsed and 'name' in parsed

    def dump(self) -> dict:
        return {
            'gameId': self.uid,
            'name': self.name,
            'firstSeenAt': self.first_seen_at.isoformat() if self.first_seen_at is not None else self.first_seen_at,
            'lastSeenAt': self.last_seen_at.isoformat(),
            'links': [link.dump() for link in self.links]
        }



from .battlelog import BattlelogServerLister
from .bfbc2 import BadCompany2ServerLister
from .gamespy import GameSpyServerLister
from .gametools import GametoolsServerLister
from .medalofhonor import MedalOfHonorServerLister
from .quake3 import Quake3ServerLister
from .unreal2 import Unreal2ServerLister
from .valve import ValveServerLister

__all__ = [
    'BadCompany2ServerLister',
    'BattlelogServerLister',
    'GameSpyServerLister',
    'GametoolsServerLister',
    'MedalOfHonorServerLister',
    'Quake3ServerLister',
    'Unreal2ServerLister',
    'ValveServerLister'
]



import json
from datetime import datetime, timedelta
from typing import Dict, Optional

from GameserverLister.common.constants import UNIX_EPOCH_START


class WebLink:
    site: str
    url: str
    official: bool
    as_of: datetime

    def __init__(self, site: str, url: str, official: bool, as_of: datetime = datetime.now().astimezone()):
        self.site = site
        self.url = url
        self.official = official
        self.as_of = as_of

    def is_expired(self, expired_ttl: float) -> bool:
        return datetime.now().astimezone() > self.as_of + timedelta(hours=expired_ttl)

    def update(self, updated: 'WebLink') -> None:
        self.url = updated.url
        self.official = updated.official
        self.as_of = updated.as_of

    @staticmethod
    def load(parsed: dict) -> 'WebLink':
        as_of = datetime.fromisoformat(parsed['asOf']) \
            if parsed.get('asOf') is not None else UNIX_EPOCH_START
        return WebLink(
            parsed['site'],
            parsed['url'],
            parsed['official'],
            as_of
        )

    @staticmethod
    def is_json_repr(parsed: dict) -> bool:
        return 'site' in parsed and 'url' in parsed and 'official' in parsed

    def dump(self) -> dict:
        return {
            'site': self.site,
            'url': self.url,
            'official': self.official,
            'asOf': self.as_of.isoformat()
        }

    def __eq__(self, other):
        return isinstance(other, WebLink) and \
               other.site == self.site and \
               other.url == self.url and \
               other.official == self.official and \
               other.as_of == self.as_of

    def __iter__(self):
        yield from self.dump().items()

    def __str__(self):
        return json.dumps(dict(self))

    def __repr__(self):
        return self.__str__()


class WebLinkTemplate:
    site: str
    url_template: str
    official: bool

    def __init__(self, site: str, url_template: str, official: bool):
        self.site = site
        self.url_template = url_template
        self.official = official

    def render(self, game: str, uid: str, ip: Optional[str] = None, port: Optional[int] = None) -> WebLink:
        return WebLink(
            self.site,
            self.url_template.format(game, uid, ip, port),
            self.official
        )


"""
For URL templates:
0: game name/key
1: server uid
2: server ip
3: server port
"""
WEB_LINK_TEMPLATES: Dict[str, WebLinkTemplate] = {
    'arena.sh': WebLinkTemplate(
        'arena.sh',
        'https://arena.sh/game/{2}:{3}/',
        False
    ),
    'battlefieldtracker': WebLinkTemplate(
        'battlefieldtracker.com',
        'https://battlefieldtracker.com/bf1/servers/pc/{1}',
        False
    ),
    'battlelog': WebLinkTemplate(
        'battlelog.com',
        'https://battlelog.battlefield.com/{0}/servers/show/pc/{1}',
        True
    ),
    'bf2.tv': WebLinkTemplate(
        'bf2.tv',
        'https://bf2.tv/servers/{2}:{3}',
        False
    ),
    'bf2hub': WebLinkTemplate(
        'bf2hub.com',
        'https://www.bf2hub.com/server/{2}:{3}/',
        True
    ),
    'cod.pm': WebLinkTemplate(
      'cod.pm',
      'https://cod.pm/server/{2}/{3}',
      False
    ),
    # deathmask.net shows servers from their own as well as other masters,
    # so they are not the official source for all servers
    'deathmask.net-official': WebLinkTemplate(
        'deathmask.net',
        'https://dpmaster.deathmask.net/?game={0}&server={2}:{3}',
        True
    ),
    'deathmask.net-unofficial': WebLinkTemplate(
        'deathmask.net',
        'https://dpmaster.deathmask.net/?game={0}&server={2}:{3}',
        False
    ),
    'gametools': WebLinkTemplate(
        'gametools.network',
        'https://gametools.network/servers/{0}/gameid/{1}/pc',
        False
    ),
    'swat4stats.com': WebLinkTemplate(
        'swat4stats.com',
        'https://swat4stats.com/servers/{2}:{3}/',
        False
    )
}



import ipaddress
import json
import logging
from typing import Callable, List

import gevent.subprocess
from nslookup import Nslookup

from GameserverLister.common.servers import FrostbiteServer, BadCompany2Server
from GameserverLister.common.types import GamespyGame


def find_query_port(gamedig_path: str, game: str, server: FrostbiteServer, ports_to_try: list, validator: Callable) -> int:
    query_port = -1

    # Add current query port add index 0 if valid
    if server.query_port != -1:
        ports_to_try.insert(0, server.query_port)
    # Try all unique ports
    for port_to_try in list(set(ports_to_try)):
        if not is_valid_port(port_to_try):
            logging.warning(f'Skipping query port to try which is outside of valid port range ({port_to_try})')
            continue

        gamedig_result = gevent.subprocess.run(
            args=[gamedig_path, '--type', game, f'{server.ip}:{port_to_try}',
                  '--maxAttempts 2', '--socketTimeout 2000', '--givenPortOnly'],
            capture_output=True
        )

        # Make sure gamedig did not log any errors to stderr and has some output in stdout
        if len(gamedig_result.stderr) > 0 or len(gamedig_result.stdout) == 0:
            continue

        # Try to parse JSON returned by gamedig
        try:
            parsed_result = json.loads(gamedig_result.stdout)
        except json.JSONDecodeError as e:
            logging.debug(e)
            logging.error('Failed to parse gamedig command output')
            continue

        # Stop searching if query was successful and response came from the correct server
        # (some servers run on the same IP, so make sure ip and game_port match)
        if not parsed_result.get('error', '').startswith('Failed all') and \
                validator(server, port_to_try, parsed_result):
            query_port = port_to_try
            break

    return query_port


def resolve_host(host: str) -> List[str]:
    if is_valid_ip(host):
        return [host]

    looker_upper = Nslookup()
    dns_result = looker_upper.dns_lookup(host)

    return dns_result.answer


def is_valid_ip(ip: str) -> bool:
    try:
        ipaddress.ip_address(ip)
        return True
    except ValueError:
        return False


def is_valid_public_ip(ip: str) -> bool:
    try:
        ip_address = ipaddress.ip_address(ip)
        return ip_address.is_global
    except ValueError:
        return False


def is_valid_port(port: int) -> bool:
    return 0 < port < 65536


def battlelog_server_validator(server: FrostbiteServer, used_query_port: int, parsed_result: dict) -> bool:
    return parsed_result.get('connect') == f'{server.ip}:{server.game_port}'


def mohwf_server_validator(server: FrostbiteServer, used_query_port: int, parsed_result: dict) -> bool:
    return parsed_result.get('connect') == f'{server.ip}:{used_query_port}'


def bfbc2_server_validator(server: BadCompany2Server, used_query_port: int, parsed_result: dict) -> bool:
    return battlelog_server_validator(server, used_query_port, parsed_result) or \
           parsed_result.get('connect') == f'0.0.0.0:{server.game_port}' or \
           parsed_result.get('name') == server.name


def is_server_for_gamespy_game(game: GamespyGame, game_name: str, parsed_result: dict) -> bool:
    """
    Check if a GameSpy query result matches key contents/structure we expect for a server of the given game
    :param game: Game the server should be from/for
    :param game_name: Name of the game as referenced by gslist
    :param parsed_result: Parsed result of a gslist GameSpy query against the server
    :return: True, if the results matches expected key content/structure, else false
    """
    if game is GamespyGame.BFVietnam:
        # Battlefield Vietnam does not reliably contain the gamename anywhere, but as some quite unique keys
        return 'allow_nose_cam' in parsed_result and 'name_tag_distance_scope' in parsed_result and \
               'soldier_friendly_fire_on_splash' in parsed_result and 'all_active_mods' in parsed_result
    elif game is GamespyGame.Crysis:
        # Crysis uses the same keys as Crysiswars, but the "gamename" key is missing
        return 'voicecomm' in parsed_result and 'dx10' in parsed_result and \
               'gamepadsonly' in parsed_result and 'gamename' not in parsed_result
    elif game is GamespyGame.Vietcong:
        # Vietcong uses many of the same keys as Vietcong 2, but the "extinfo" key is missing (amongst others)
        return 'uver' in parsed_result and 'dedic' in parsed_result and 'extinfo' not in parsed_result
    elif game is GamespyGame.Vietcong2:
        return 'uver' in parsed_result and 'dedic' in parsed_result and 'extinfo' in parsed_result
    elif game is GamespyGame.FH2:
        # Check mod value, since Forgotten Hope 2 is technically a Battlefield 2 mod
        return parsed_result.get('gamename') == game_name and parsed_result.get('gamevariant') == 'fh2'
    elif game is GamespyGame.SWAT4:
        # SWAT 4 has a very limited set of keys, so we need to look at values
        return parsed_result.get('gamevariant') in ['SWAT 4', 'SWAT 4X', 'SEF', 'FR']
    elif game is GamespyGame.UT3:
        # UT3 has some *very* unique keys, see https://github.com/gamedig/node-gamedig/blob/master/protocols/ut3.js#L7
        return 'p1073741825' in parsed_result and 'p1073741826' in parsed_result
    else:
        return parsed_result.get('gamename', '').lower() == game_name


def guid_from_ip_port(ip: str, port: str) -> str:
    int_port = int(port)
    guid = '-'.join([f'{int((pow(int(octet) + 2, 2)*pow(int_port, 2))/(int_port*8)):0>x}' for
                     (index, octet) in enumerate(ip.split('.'))])

    return guid



from typing import Dict

from GameserverLister.common.types import ValvePrincipal, ValvePrincipalConfig, ValveGame, ValveGameConfig

VALVE_PRINCIPAL_CONFIGS: Dict[ValvePrincipal, ValvePrincipalConfig] = {
    ValvePrincipal.VALVE: ValvePrincipalConfig(
        hostname='hl2master.steampowered.com',
        port=27011
    )
}
VALVE_GAME_CONFIGS: Dict[ValveGame, ValveGameConfig] = {
    ValveGame.AmericasArmyProvingGrounds: ValveGameConfig(
        app_id=203290,
        principals=[
            ValvePrincipal.VALVE
        ],
        distinct_query_port=True
    ),
    ValveGame.ARKSurvivalEvolved: ValveGameConfig(
        app_id=346110,
        principals=[
            ValvePrincipal.VALVE
        ]
    ),
    ValveGame.Arma2: ValveGameConfig(
        app_id=33930,
        principals=[
            ValvePrincipal.VALVE
        ],
        distinct_query_port=True
    ),
    ValveGame.Arma3: ValveGameConfig(
        app_id=107410,
        principals=[
            ValvePrincipal.VALVE
        ],
        distinct_query_port=True
    ),
    ValveGame.CounterStrike: ValveGameConfig(
        app_id=10,
        principals=[
            ValvePrincipal.VALVE
        ]
    ),
    ValveGame.CounterStrikeConditionZero: ValveGameConfig(
        app_id=80,
        principals=[
            ValvePrincipal.VALVE
        ]
    ),
    ValveGame.CounterStrikeSource: ValveGameConfig(
        app_id=240,
        principals=[
            ValvePrincipal.VALVE
        ]
    ),
    ValveGame.CounterStrikeGlobalOffensive: ValveGameConfig(
        app_id=730,
        principals=[
            ValvePrincipal.VALVE
        ]
    ),
    ValveGame.DayZ: ValveGameConfig(
        app_id=221100,
        principals=[
            ValvePrincipal.VALVE
        ],
        distinct_query_port=True
    ),
    ValveGame.DayZMod: ValveGameConfig(
        app_id=224580,
        principals=[
            ValvePrincipal.VALVE
        ],
        distinct_query_port=True
    ),
    ValveGame.DoD: ValveGameConfig(
        app_id=30,
        principals=[
            ValvePrincipal.VALVE
        ]
    ),
    ValveGame.DoDS: ValveGameConfig(
        app_id=300,
        principals=[
            ValvePrincipal.VALVE
        ]
    ),
    ValveGame.GarrysMod: ValveGameConfig(
        app_id=4000,
        principals=[
            ValvePrincipal.VALVE
        ]
    ),
    ValveGame.Insurgency: ValveGameConfig(
        app_id=222880,
        principals=[
            ValvePrincipal.VALVE
        ]
    ),
    ValveGame.InsurgencySandstorm: ValveGameConfig(
        app_id=581320,
        principals=[
            ValvePrincipal.VALVE
        ],
        distinct_query_port=True
    ),
    ValveGame.Left4Dead: ValveGameConfig(
        app_id=500,
        principals=[
            ValvePrincipal.VALVE
        ]
    ),
    ValveGame.Left4Dead2: ValveGameConfig(
        app_id=550,
        principals=[
            ValvePrincipal.VALVE
        ]
    ),
    ValveGame.RS2: ValveGameConfig(
        app_id=418460,
        principals=[
            ValvePrincipal.VALVE
        ]
    ),
    ValveGame.Rust: ValveGameConfig(
        app_id=252490,
        principals=[
            ValvePrincipal.VALVE
        ]
    ),
    ValveGame.SevenD2D: ValveGameConfig(
        app_id=251570,
        principals=[
            ValvePrincipal.VALVE
        ]
    ),
    ValveGame.Squad: ValveGameConfig(
        app_id=393380,
        principals=[
            ValvePrincipal.VALVE
        ],
        distinct_query_port=True
    ),
    ValveGame.TFC: ValveGameConfig(
        app_id=20,
        principals=[
            ValvePrincipal.VALVE
        ]
    ),
    ValveGame.TF2: ValveGameConfig(
        app_id=440,
        principals=[
            ValvePrincipal.VALVE
        ]
    )
}



import logging
from typing import List, Tuple, Optional

import pyvpsq

from GameserverLister.common.helpers import is_valid_public_ip, is_valid_port, guid_from_ip_port
from GameserverLister.common.servers import ClassicServer, ViaStatus
from GameserverLister.common.types import ValveGame, ValvePrincipal, ValveGameConfig
from GameserverLister.games.valve import VALVE_PRINCIPAL_CONFIGS, VALVE_GAME_CONFIGS
from GameserverLister.listers.common import ServerLister


class ValveServerLister(ServerLister):
    game: ValveGame
    servers: List[ClassicServer]
    principal: ValvePrincipal
    config: ValveGameConfig

    principal_timeout: float
    filters: str
    max_pages: int

    def __init__(
            self,
            game: ValveGame,
            principal: ValvePrincipal,
            principal_timeout: float,
            filters: str,
            max_pages: int,
            expired_ttl: float,
            recover: bool,
            add_links: bool,
            list_dir: str
    ):
        super().__init__(game, ClassicServer, expired_ttl, recover, add_links, list_dir)
        self.principal = principal
        self.config = VALVE_GAME_CONFIGS[self.game]
        self.principal_timeout = principal_timeout
        self.filters = filters
        self.max_pages = max_pages

    def update_server_list(self):
        principal_config = VALVE_PRINCIPAL_CONFIGS[self.principal]
        principal = pyvpsq.PrincipalServer(
            principal_config.hostname,
            principal_config.port,
            timeout=self.principal_timeout
        )

        found_servers = []
        # Try to reduce the consecutive number of requests by iterating over regions
        for region in pyvpsq.Region:
            for raw_server in self.get_servers(principal, self.config.app_id, region, self.filters, self.max_pages):
                if not is_valid_public_ip(raw_server.ip) or not is_valid_port(raw_server.query_port):
                    logging.warning(
                        f'Principal returned invalid server entry '
                        f'({raw_server.ip}:{raw_server.query_port}), skipping it'
                    )
                    continue

                via = ViaStatus(self.principal)
                found_server = ClassicServer(
                    guid_from_ip_port(raw_server.ip, str(raw_server.query_port)),
                    raw_server.ip,
                    raw_server.query_port,
                    via
                )

                if found_server not in found_servers:
                    if self.add_links:
                        game_port = self.get_server_game_port(found_server)
                        if game_port is not None:
                            found_server.add_links(self.build_server_links(
                                found_server.uid,
                                found_server.ip,
                                game_port
                            ))
                    found_servers.append(found_server)

        self.add_update_servers(found_servers)

    @staticmethod
    def get_servers(
            principal: pyvpsq.PrincipalServer,
            app_id: int,
            region: pyvpsq.Region,
            filters: str,
            max_pages: int
    ) -> List[pyvpsq.Server]:
        servers = []
        try:
            for server in principal.get_servers(fr'\appid\{app_id}{filters}', region, max_pages):
                servers.append(server)
        except pyvpsq.TimeoutError:
            logging.error('Principal server query timed out')
        except pyvpsq.Error:
            logging.error('Failed to query principal server')

        return servers

    def get_server_game_port(self, server: ClassicServer) -> Optional[int]:
        if not self.config.distinct_query_port:
            return server.query_port

        responded, info = self.query_server(server)
        if responded and info.game_port is not None:
            return info.game_port

    def check_if_server_still_exists(self, server: ClassicServer, checks_since_last_ok: int) -> Tuple[bool, bool, int]:
        found, _ = self.query_server(server)
        return True, found, checks_since_last_ok

    @staticmethod
    def query_server(server: ClassicServer) -> Tuple[bool, Optional[pyvpsq.ServerInfo]]:
        try:
            info = pyvpsq.Server(server.ip, server.query_port).get_info()
            return True, info
        except pyvpsq.Error as e:
            logging.debug(e)
            logging.debug(f'Failed to query server {server.uid}')
            return False, None



import logging
import os
import subprocess
import sys
from typing import List, Tuple, Optional, Union

from GameserverLister.common.helpers import is_valid_public_ip, is_valid_port, guid_from_ip_port, \
    is_server_for_gamespy_game, resolve_host
from GameserverLister.common.servers import ClassicServer, ViaStatus
from GameserverLister.common.types import GamespyGame, GamespyPrincipal, GamespyGameConfig
from GameserverLister.common.weblinks import WebLink, WEB_LINK_TEMPLATES
from GameserverLister.games.gamespy import GAMESPY_PRINCIPAL_CONFIGS, GAMESPY_GAME_CONFIGS
from .common import ServerLister


class GameSpyServerLister(ServerLister):
    game: GamespyGame
    servers: List[ClassicServer]
    principal: GamespyPrincipal
    config: GamespyGameConfig
    gslist_bin_path: str
    gslist_filter: str
    gslist_super_query: bool
    gslist_timeout: int
    verify: bool

    def __init__(
            self,
            game: GamespyGame,
            principal: GamespyPrincipal,
            gslist_bin_path: str,
            gslist_filter: str,
            gslist_super_query: bool,
            gslist_timeout: int,
            verify: bool,
            expired_ttl: float,
            recover: bool,
            add_links: bool,
            list_dir: str
    ):
        super().__init__(game, ClassicServer, expired_ttl, recover, add_links, list_dir)
        self.principal = principal.lower()
        self.config = GAMESPY_GAME_CONFIGS[self.game]
        self.gslist_bin_path = gslist_bin_path
        self.gslist_filter = gslist_filter
        self.gslist_super_query = gslist_super_query
        self.gslist_timeout = gslist_timeout
        self.verify = verify

    def update_server_list(self):
        raw_servers = self.get_servers()
        found_servers = []
        for ip, query_port in raw_servers:
            if not is_valid_public_ip(ip) or not is_valid_port(int(query_port)):
                logging.warning(f'Principal returned invalid server entry ({ip}:{query_port}), skipping it')
                continue

            via = ViaStatus(self.principal)
            found_server = ClassicServer(
                guid_from_ip_port(ip, str(query_port)),
                ip,
                query_port,
                via
            )

            if self.verify or self.add_links:
                # Attempt to query server in order to verify is a server for the current game
                # (some principals return servers for other games than what we queried)
                logging.debug(f'Querying server {found_server.uid}/{found_server.ip}:{found_server.query_port}')
                responded, query_response = self.query_server(found_server)
                logging.debug(f'Query {"was successful" if responded else "did not receive a response"}')

                if responded and self.verify and \
                        not is_server_for_gamespy_game(self.game, self.config.game_name, query_response):
                    logging.warning(f'Server does not seem to be a {self.game} server, ignoring it '
                                    f'({found_server.ip}:{found_server.query_port})')
                    continue

                if responded and self.add_links:
                    if query_response.get('hostport', '').isnumeric():
                        found_server.add_links(self.build_server_links(
                            found_server.uid,
                            found_server.ip,
                            int(query_response['hostport'])
                        ))
                    elif 'hostport' in query_response:
                        logging.warning(f'Server returned an invalid hostport (\'{query_response["hostport"]}\', ' 
                                        f'not adding links ({found_server.ip}:{found_server.query_port})')

            found_servers.append(found_server)

        self.add_update_servers(found_servers)

    def get_servers(self) -> List[Tuple[str, int]]:
        principal = GAMESPY_PRINCIPAL_CONFIGS[self.principal]
        # Format hostname using game name (following old GameSpy format [game].master.gamespy.com)
        hostname = principal.hostname.format(self.config.game_name)
        # Combine game port and principal-specific port offset (defaults to an offset of 0)
        port = self.config.port + principal.get_port_offset()

        # Manually look up hostname to be able to spread retried across servers
        ips = resolve_host(hostname)
        if len(ips) == 0:
            logging.error(f'DNS lookup for {hostname} failed, exiting')
            sys.exit(1)

        # Run gslist and capture output
        command_ok = False
        tries = 0
        max_tries = 3
        gslist_result = None
        while not command_ok and tries < max_tries:
            # Alternate between first and last found A record
            ip = ips[0] if tries % 2 == 0 else ips[-1]
            try:
                logging.info(f'Running gslist command against {ip}')
                command = [self.gslist_bin_path, '-n', self.config.game_name, '-x',
                           f'{ip}:{port}', '-Y', self.config.game_name, self.config.game_key,
                           '-t', str(self.config.enc_type), '-f', f'{self.gslist_filter}', '-o', '1']
                timeout = self.gslist_timeout
                # Some principals do not respond with the default query list type byte (1),
                # so we need to explicitly set a different type byte
                if self.config.list_type is not None:
                    command.extend(['-T', str(self.config.list_type)])
                # Some principals do not respond unless an info query is sent (e.g. FH2 principal)
                if self.config.info_query is not None:
                    command.extend(['-X', self.config.info_query])
                # Add super query argument if requested
                if self.gslist_super_query:
                    command.extend(['-Q', str(self.config.query_type)])
                    # Extend timeout to account for server queries
                    timeout += 10
                gslist_result = subprocess.run(
                    command, capture_output=True,
                    cwd=self.server_list_dir_path, timeout=timeout
                )
                command_ok = True
            except subprocess.TimeoutExpired as e:
                logging.debug(e)
                logging.error(f'gslist timed out, try {tries + 1}/{max_tries}')
                tries += 1

        # Make sure any server were found
        # (gslist sends all output to stderr so check there)
        if gslist_result is None or 'servers found' not in str(gslist_result.stderr):
            logging.error('gslist could not retrieve any servers, exiting')
            sys.exit(1)

        # Read gslist output file
        logging.info('Reading gslist output file')
        with open(os.path.join(self.server_list_dir_path, f'{self.config.game_name}.gsl'), 'r') as gslist_file:
            raw_server_list = gslist_file.read()

        # Parse server list
        # List format: [ip-address]:[port]
        logging.info(f'Parsing server list{" and verifying servers" if self.verify else ""}')
        servers: List[Tuple[str, int]] = []
        for line in raw_server_list.splitlines():
            connect, *_ = line.split(' ', 1)
            ip, query_port = connect.strip().split(':', 1)
            servers.append((ip, int(query_port)))

        return servers

    def check_if_server_still_exists(self, server: ClassicServer, checks_since_last_ok: int) -> Tuple[bool, bool, int]:
        # Since we query the server directly, there is no way of handling HTTP server errors differently then
        # actually failed checks, so even if the query fails, we have to treat it as "check ok"
        check_ok = True
        responded, query_response = self.query_server(server)
        # Treat as server for game if verify is turned off
        server_for_game = not self.verify or is_server_for_gamespy_game(self.game, self.config.game_name, query_response)
        found = responded and server_for_game
        if responded and not server_for_game:
            logging.warning(f'Server {server.uid} does not seem to be a {self.game} server, treating as not found')

        return check_ok, found, checks_since_last_ok

    def build_server_links(
            self,
            uid: str,
            ip: Optional[str] = None,
            port: Optional[int] = None
    ) -> Union[List[WebLink], WebLink]:
        template_refs = self.config.link_template_refs if self.config.link_template_refs is not None else {}
        # Add principal-scoped links first, then add game-scoped links
        templates = [
            *[WEB_LINK_TEMPLATES.get(ref) for ref in template_refs.get(self.principal, [])],
            *[WEB_LINK_TEMPLATES.get(ref) for ref in template_refs.get('_any', [])]
        ]

        links = []
        for template in templates:
            links.append(template.render(self.game, uid, ip=ip, port=port))

        return links

    def query_server(self, server: ClassicServer) -> Tuple[bool, dict]:
        try:
            command = [self.gslist_bin_path, '-d', str(self.config.query_type), server.ip, str(server.query_port), '-0']
            # Timeout should never fire since gslist uses about a three-second timeout for the query
            gslist_result = subprocess.run(command, capture_output=True, timeout=self.gslist_timeout)

            # gslist will simply return an empty byte string (b'') if the server could not be queried
            if gslist_result.stdout != b'' and b'error' not in gslist_result.stderr.lower():
                parsed = {}
                for line in gslist_result.stdout.decode('latin1').strip('\n').split('\n'):
                    elements = line.lstrip().split(' ', 1)
                    if len(elements) != 2:
                        continue
                    key, value = elements
                    parsed[key.lower()] = value
                return True, parsed
        except subprocess.TimeoutExpired as e:
            logging.debug(e)
            logging.error(f'Failed to query server {server.uid} for expiration check')

        return False, {}



import logging
import sys

import click

from GameserverLister.commands.options import common, queryport
from GameserverLister.common.logger import logger
from GameserverLister.listers import BadCompany2ServerLister


@click.command
@click.option(
    '-t',
    '--timeout',
    type=int,
    default=10,
    help='Timeout to use for server list retrieval request'
)
@queryport.find
@queryport.gamedig_bin
@queryport.gamedig_concurrency
@common.expired_ttl
@common.list_dir
@common.recover
@common.add_links
@common.debug
def run(
        timeout: int,
        find_query_port: bool,
        gamedig_bin: str,
        gamedig_concurrency: int,
        expired_ttl: int,
        recover: bool,
        add_links: bool,
        list_dir: str,
        debug: bool
):
    logging.basicConfig(level=logging.DEBUG if debug else logging.INFO, stream=sys.stdout,
                        format='%(asctime)s %(levelname)-8s %(message)s')
    logger.info('Listing servers for bfbc2')

    lister = BadCompany2ServerLister(
        expired_ttl,
        recover,
        add_links,
        list_dir,
        timeout
    )

    before = len(lister.servers)
    lister.update_server_list()

    if find_query_port:
        lister.find_query_ports(gamedig_bin, gamedig_concurrency, expired_ttl)

    removed, recovered = lister.remove_expired_servers()
    lister.write_to_file()

    logger.info(f'Server list updated ('
                f'total: {len(lister.servers)}, '
                f'added: {len(lister.servers) + removed - before}, '
                f'removed: {removed}, '
                f'recovered: {recovered})')



import logging
from datetime import datetime
from typing import List, Tuple, Optional, Union

import requests

from GameserverLister.common.helpers import mohwf_server_validator, battlelog_server_validator
from GameserverLister.common.servers import FrostbiteServer
from GameserverLister.common.types import BattlelogGame
from GameserverLister.common.weblinks import WEB_LINK_TEMPLATES, WebLink
from GameserverLister.games.battlelog import BATTLELOG_GAME_BASE_URIS
from .common import HttpServerLister, FrostbiteServerLister


class BattlelogServerLister(HttpServerLister, FrostbiteServerLister):
    game: BattlelogGame

    def __init__(
            self,
            game: BattlelogGame,
            page_limit: int,
            expired_ttl: float,
            recover: bool,
            add_links: bool,
            list_dir: str,
            sleep: float,
            max_attempts: int,
            proxy: str = None
    ):
        super().__init__(game, FrostbiteServer, page_limit, 60, expired_ttl, recover, add_links, list_dir, sleep, max_attempts)
        # Medal of Honor: Warfighter servers return the query port as part of the connect string, not the game port
        # => use different validator
        if self.game is BattlelogGame.MOHWF:
            self.server_validator = mohwf_server_validator
        else:
            self.server_validator = battlelog_server_validator

        # Set up headers
        self.session.headers = {
            'X-Requested-With': 'XMLHttpRequest'
        }
        # Set up proxy if given
        if proxy is not None:
            # All requests are sent via https, so just set up https proxy
            self.session.proxies = {
                'https': proxy
            }

    def get_server_list_url(self, per_page: int) -> str:
        return f'{BATTLELOG_GAME_BASE_URIS[self.game]}?count={per_page}&offset=0'

    def add_page_found_servers(self, found_servers: List[FrostbiteServer], page_response_data: dict) -> List[FrostbiteServer]:
        for server in page_response_data['data']:
            found_server = FrostbiteServer(
                server['guid'],
                server['name'],
                server['ip'],
                server['port'],
            )

            if self.add_links:
                found_server.add_links(self.build_server_links(found_server.uid))
                # Gametools uses the gameid for BF4 server URLs, so add that separately
                if self.game is BattlelogGame.BF4:
                    found_server.add_links(WEB_LINK_TEMPLATES['gametools'].render(self.game, server['gameId']))

            # Add non-private servers (servers with an IP) that are new
            server_uids = [s.uid for s in found_servers]
            if len(found_server.ip) > 0 and found_server.uid not in server_uids:
                logging.debug(f'Got new server {found_server.uid}, adding it')
                found_servers.append(found_server)
            elif len(found_server.ip) > 0:
                logging.debug(f'Got duplicate server {found_server.uid}, updating last seen at')
                index = server_uids.index(found_server.uid)
                found_servers[index].last_seen_at = datetime.now().astimezone()
            else:
                logging.debug(f'Got private server {found_server.uid}, ignoring it')

        return found_servers

    def check_if_server_still_exists(self, server: FrostbiteServer, checks_since_last_ok: int) -> Tuple[bool, bool, int]:
        check_ok = True
        found = False
        try:
            response = self.session.get(f'https://battlelog.battlefield.com/{self.game}/'
                                        f'servers/show/pc/{server.uid}?json=1',
                                        timeout=self.request_timeout)
            if response.status_code == 200:
                # Server was found on Battlelog => make sure it is still public
                # TODO use server validator here
                parsed = response.json()
                found = parsed['message']['SERVER_INFO']['ip'] != ''
            elif response.status_code == 422:
                # Battlelog responded with 422, explicitly indicating that the server was not found
                found = False
            else:
                # Battlelog responded with some other status code (rate limit 403 for example)
                check_ok = False
            # Reset requests since last ok counter if server returned info/not found,
            # else increase counter and sleep
            if check_ok:
                checks_since_last_ok = 0
            else:
                checks_since_last_ok += 1
        except requests.exceptions.RequestException as e:
            logging.debug(e)
            logging.error(f'Failed to fetch server {server.uid} for expiration check')
            check_ok = False

        return check_ok, found, checks_since_last_ok

    def build_server_links(
            self,
            uid: str,
            ip: Optional[str] = None,
            port: Optional[int] = None
    ) -> Union[List[WebLink], WebLink]:
        # Medal of Honor: Warfighter (mohwf) is just called "mohw" on Battlelog
        game = 'mohw' if self.game is BattlelogGame.MOHWF else self.game
        links = [WEB_LINK_TEMPLATES['battlelog'].render(game, uid)]

        # Gametools uses the guid as the "gameid" for BF3 and BFH, so we can add links for those
        # (BF4 uses the real gameid, so we need to handle those links separately)
        if self.game in ['bf3', 'bfh']:
            links.append(WEB_LINK_TEMPLATES['gametools'].render(self.game, uid))

        return links

    def build_port_to_try_list(self, game_port: int) -> list:
        """
                Order of ports to try:
                1. default query port
                2. game port + default port offset (mirror gamedig behavior)
                3. game port (some servers use same port for game + query)
                4. game port + 100 (nitrado)
                5. game port + 5 (several hosters)
                6. 48888 (gamed)
                7. game port + 50
                8. game port + 6 (i3D)
                9. game port + 8 (i3D)
                10. game port + 15 (i3D)
                11. game port - 5 (i3D)
                12. game port - 15 (i3D)
                13. game port - 23000 (G4G.pl)
                """
        ports_to_try = [47200, game_port + 22000, game_port, game_port + 100, game_port + 5, game_port + 1,
                        48888, game_port + 6, game_port + 8, game_port + 15, game_port - 5, game_port - 15,
                        game_port - 23000]

        return ports_to_try



from dataclasses import dataclass
from enum import Enum
from typing import List, Optional, Dict, Union


# https://stackoverflow.com/a/54919285/9395553
class ExtendedEnum(Enum):
    @classmethod
    def list(cls):
        return list(map(lambda c: c.value, cls))


class Principal(str, ExtendedEnum):
    pass


class GamespyPrincipal(Principal):
    TripleThreeNetworks_com_1 = '333networks.com-1'
    TripleThreeNetworks_com_2 = '333networks.com-2'
    BF1942_org = 'bf1942.org'
    BF2Hub_com = 'bf2hub.com'
    Crymp_net = 'crymp.net'
    Errorist_eu = 'errorist.eu'
    FH2_dev = 'fh2.dev'
    Jedi95_us = 'jedi95.us'
    Newbiesplayground_net = 'newbiesplayground.net'
    NightfirePC_com = 'nightfirepc.com'
    NovGames_ru = 'novgames.ru'
    OldUnreal_com_1 = 'oldunreal.com-1'
    OldUnreal_com_2 = 'oldunreal.com-2'
    OpenSpy_net = 'openspy.net'
    PhoenixNetwork_net = 'phoenixnetwork.net'
    Play2142_ru = 'play2142.ru'
    PlayBF2_ru = 'playbf2.ru'
    Qtracker_com = 'qtracker.com'
    SWAT4Stats_com = 'swat4stats.com'
    Vietcong_tk = 'vietcong.tk'
    Vietcong1_eu = 'vietcong1.eu'


class ValvePrincipal(Principal):
    VALVE = 'valve'


@dataclass
class GamespyPrincipalConfig:
    hostname: str
    port_offset: Optional[int] = None

    def get_port_offset(self) -> int:
        if self.port_offset is None:
            return 0
        return self.port_offset


@dataclass
class ValvePrincipalConfig:
    hostname: str
    port: int


class Game(str, ExtendedEnum):
    pass


class GamespyGame(Game):
    BF1942 = 'bf1942'
    BFVietnam = 'bfvietnam'
    BF2 = 'bf2'
    FH2 = 'fh2'
    BF2142 = 'bf2142'
    Crysis = 'crysis'
    CrysisWars = 'crysiswars'
    DeusEx = 'deusex'
    DukeNukemForever = 'dnf'
    JBNightfire = 'jbnightfire'
    Paraworld = 'paraworld'
    Postal2 = 'postal2'
    Rune = 'rune'
    SeriousSam = 'serioussam'
    SeriousSamSE = 'serioussamse'
    SWAT4 = 'swat4'
    Unreal = 'unreal'
    UT = 'ut'
    UT3 = 'ut3'
    Vietcong = 'vietcong'
    Vietcong2 = 'vietcong2'
    WheelOfTime = 'wot'


class Quake3Game(Game):
    CoD = 'cod'
    CoDUO = 'coduo'
    CoD2 = 'cod2'
    CoD4 = 'cod4'
    CoD4X = 'cod4x'
    Nexuiz = 'nexuiz'
    OpenArena = 'openarena'
    Q3Rally = 'q3rally'
    Quake = 'quake'
    Quake3Arena = 'quake3arena'
    RTCW = 'rtcw'
    SOF2 = 'sof2'
    SWJKJA = 'swjkja'
    SWJKJO = 'swjkjo'
    Tremulous = 'tremulous'
    UrbanTerror = 'urbanterror'
    Warfork = 'warfork'
    Warsow = 'warsow'
    WolfensteinET = 'wolfensteinet'
    Xonotic = 'xonotic'


class MedalOfHonorGame(Game):
    AA = 'mohaa'
    BT = 'mohbt'
    PA = 'mohpa'
    SH = 'mohsh'


class Unreal2Game(Game):
    UT2003 = 'ut2003'
    UT2004 = 'ut2004'


class ValveGame(Game):
    AmericasArmyProvingGrounds = 'aapg'
    ARKSurvivalEvolved = 'arkse'
    Arma2 = 'arma2'
    Arma3 = 'arma3'
    CounterStrike  = 'cs'
    CounterStrikeConditionZero = 'cscz'
    CounterStrikeSource = 'css'
    CounterStrikeGlobalOffensive = 'csgo'
    DayZ = 'dayz'
    DayZMod = 'dayzmod'
    DoD = 'dod'
    DoDS = 'dods'
    GarrysMod = 'gmod'
    Insurgency = 'insurgency'
    InsurgencySandstorm = 'insurgency-sandstorm'
    Left4Dead = 'left4dead'
    Left4Dead2 = 'left4dead2'
    RS2 = 'rs2'
    Rust = 'rust'
    SevenD2D = '7d2d'
    Squad = 'squad'
    TFC = 'tfc'
    TF2 = 'tf2'


class TheaterGame(Game):
    BFBC2 = 'bfbc2'


class BattlelogGame(Game):
    BF3 = 'bf3'
    BF4 = 'bf4'
    BFH = 'bfh'
    MOHWF = 'mohwf'


class GametoolsGame(Game):
    BF1 = 'bf1'
    BFV = 'bfv'


@dataclass
class GamespyGameConfig:
    game_name: str
    game_key: str
    enc_type: int
    query_type: int
    port: int
    principals: List[GamespyPrincipal]
    list_type: Optional[int] = None
    info_query: Optional[str] = None
    link_template_refs: Optional[Dict[Union[str, GamespyPrincipal], List[str]]] = None


@dataclass
class ValveGameConfig:
    app_id: int
    principals: List[ValvePrincipal]
    distinct_query_port: bool = False
    link_template_refs: Optional[Dict[Union[str, ValvePrincipal], List[str]]] = None



import click

page_limit = click.option(
    '-p',
    '--page-limit',
    type=int,
    default=10,
    help='Number of pages to get after retrieving the last unique server'
)
sleep = click.option(
    '--sleep',
    type=float,
    default=0,
    help='Number of seconds to sleep between requests'
)
max_attempts = click.option(
    '--max-attempts',
    type=int,
    default=3,
    help='Max number of attempts for fetching a page of servers'
)
proxy = click.option(
    '--proxy',
    type=str,
    help='Proxy to use for requests (format: [protocol]://[username]:[password]@[hostname]:[port]'
)



import socket
from typing import Dict

from GameserverLister.common.types import Quake3Game

QUAKE3_CONFIGS: Dict[Quake3Game, dict] = {
    Quake3Game.CoD: {
        'protocols': [
            1,  # version 1.1
            2,  # version 1.2
            4,  # version 1.3
            5,  # version 1.4
            6,  # version 1.5
        ],
        'servers': {
            'activision': {
                'hostname': 'codmaster.activision.com',
                'port': 20510
            }
        },
        'linkTemplateRefs': {
            'activision': ['cod.pm']
        }
    },
    Quake3Game.CoDUO: {
        'protocols': [
            21,  # version 1.41
            22,  # version 1.51
        ],
        'servers': {
            'activision': {
                'hostname': 'coduomaster.activision.com',
                'port': 20610
            }
        },
        'linkTemplateRefs': {
            'activision': ['cod.pm']
        }
    },
    Quake3Game.CoD2: {
        'protocols': [
            115,  # version 1.0
            117,  # version 1.2
            118,  # version 1.3
        ],
        'servers': {
            'activision': {
                'hostname': 'cod2master.activision.com',
                'port': 20710
            }
        },
        'linkTemplateRefs': {
            'activision': ['cod.pm']
        }
    },
    Quake3Game.CoD4: {
        'protocols': [
            1,  # version 1.0
            6,  # version 1.7
            7,  # version 1.8
        ],
        'servers': {
            'activision': {
                'hostname': 'cod4master.activision.com',
                'port': 20810
            }
        },
        'linkTemplateRefs': {
            'activision': ['cod.pm']
        }
    },
    Quake3Game.CoD4X: {
        'protocols': [
            6  # cod4 does not support different protocols, you seem to get the same servers regardless
        ],
        'keywords': 'full empty \x00',
        'game_name': 'cod4x',
        'network_protocol': socket.SOCK_STREAM,
        'server_entry_prefix': b'\x00\x00\x00\x00\x04',
        'servers': {
            'cod4x.ovh': {
                'hostname': 'cod4master.cod4x.ovh',
                'port': 20810
            }
        }
    },
    Quake3Game.Nexuiz: {
        'protocols': [
            3
        ],
        'game_name': 'Nexuiz',
        'servers': {
            'deathmask.net': {
                'hostname': 'dpmaster.deathmask.net',
                'port': 27950
            }
        },
        'linkTemplateRefs': {
            'deathmask.net': ['deathmask.net-official']
        }
    },
    Quake3Game.OpenArena: {
        'protocols': [
            71
        ],
        'game_name': 'Quake3Arena',
        'servers': {
            'deathmask.net': {
                'hostname': 'dpmaster.deathmask.net',
                'port': 27950
            }
        },
        'linkTemplateRefs': {
            'deathmask.net': ['deathmask.net-official', 'arena.sh']
        }
    },
    Quake3Game.Q3Rally: {
        'protocols': [
            71
        ],
        'game_name': 'Q3Rally',
        'servers': {
            'deathmask.net': {
                'hostname': 'dpmaster.deathmask.net',
                'port': 27950
            }
        },
        'linkTemplateRefs': {
            'deathmask.net': ['deathmask.net-official']
        }
    },
    Quake3Game.Quake: {
        'protocols': [
            3
        ],
        'game_name': 'DarkPlaces-Quake',
        'servers': {
            'deathmask.net': {
                'hostname': 'dpmaster.deathmask.net',
                'port': 27950
            }
        },
        'linkTemplateRefs': {
            'deathmask.net': ['deathmask.net-official']
        }
    },
    Quake3Game.Quake3Arena: {
        'protocols': [
            68
        ],
        'servers': {
            'quake3arena.com': {
                'hostname': 'master.quake3arena.com',
                'port': 27950
            },
            'urbanterror.info-1': {
                'hostname': 'master.urbanterror.info',
                'port': 27900
            },
            'urbanterror.info-2': {
                'hostname': 'master2.urbanterror.info',
                'port': 27900
            },
            'excessiveplus.net': {
                'hostname': 'master0.excessiveplus.net',
                'port': 27950
            },
            'ioquake3.org': {
                'hostname': 'master.ioquake3.org',
                'port': 27950
            },
            'huxxer.de': {
                'hostname': 'master.huxxer.de',
                'port': 27950
            },
            'maverickservers.com': {
                'hostname': 'master.maverickservers.com',
                'port': 27950
            },
            'deathmask.net': {
                'hostname': 'dpmaster.deathmask.net',
                'port': 27950
            }
        }
    },
    Quake3Game.RTCW: {
        'protocols': [
            57
        ],
        'servers': {
            'idsoftware': {
                'hostname': 'wolfmaster.idsoftware.com',
                'port': 27950
            }
        }
    },
    Quake3Game.SOF2: {
        'protocols': [
            2001,  # version sof2mp-1.02t (demo)
            2002,  # version sof2mp-1.00 (full)
            2004,  # version sof2mp-1.02 ("gold")
        ],
        'servers': {
            'ravensoft': {
                'hostname': 'master.sof2.ravensoft.com',
                'port': 20110
            }
        }
    },
    Quake3Game.SWJKJA: {
        'protocols': [
            25,  # version 1.00
            26,  # version 1.01
        ],
        'servers': {
            'ravensoft': {
                'hostname': 'masterjk3.ravensoft.com',
                'port': 29060
            },
            'jkhub.org': {
                'hostname': 'master.jkhub.org',
                'port': 29060
            }
        }
    },
    Quake3Game.SWJKJO: {
        'protocols': [
            15,  # version 1.02
            16,  # version 1.04
        ],
        'servers': {
            'ravensoft': {
                'hostname': 'masterjk2.ravensoft.com',
                'port': 28060
            },
            'jkhub.org': {
                'hostname': 'master.jkhub.org',
                'port': 28060
            }
        }
    },
    Quake3Game.Tremulous: {
        'protocols': [
            69
        ],
        'servers': {
            'tremulous.net': {
                'hostname': 'master.tremulous.net',
                'port': 30710
            }
        },
        'linkTemplateRefs': {
            'tremulous.net': ['deathmask.net-unofficial']
        }
    },
    Quake3Game.UrbanTerror: {
        'protocols': [
            68
        ],
        'servers': {
            'urbanterror.info': {
                'hostname': 'master.urbanterror.info',
                'port': 27900
            }
        },
        'linkTemplateRefs': {
            'urbanterror.info': ['deathmask.net-unofficial']
        }
    },
    Quake3Game.Warfork: {
        'protocols': [
            23
        ],
        'game_name': 'Warfork',
        'servers': {
            'deathmask.net': {
                'hostname': 'dpmaster.deathmask.net',
                'port': 27950
            }
        },
        'linkTemplateRefs': {
            'deathmask.net': ['deathmask.net-official']
        }
    },
    Quake3Game.Warsow: {
        'protocols': [
            22
        ],
        'game_name': 'Warsow',
        'servers': {
            'deathmask.net': {
                'hostname': 'dpmaster.deathmask.net',
                'port': 27950
            }
        },
        'linkTemplateRefs': {
            'deathmask.net': ['deathmask.net-official', 'arena.sh']
        }
    },
    Quake3Game.WolfensteinET: {
        'protocols': [
            84
        ],
        'servers': {
            'idsoftware': {
                'hostname': 'etmaster.idsoftware.com',
                'port': 27950
            },
            'etlegacy.com': {
                'hostname': 'master.etlegacy.com',
                'port': 27950
            },
            'etmaster.net': {
                'hostname': 'master0.etmaster.net',
                'port': 27950
            }
        }
    },
    Quake3Game.Xonotic: {
        'protocols': [
            3
        ],
        'game_name': 'Xonotic',
        'servers': {
            'deathmask.net': {
                'hostname': 'dpmaster.deathmask.net',
                'port': 27950
            },
            'tchr.no': {
                'hostname': 'dpmaster.tchr.no',
                'port': 27950
            }
        },
        'linkTemplateRefs': {
            'deathmask.net': ['deathmask.net-official', 'arena.sh']
        }
    }
}



import logging
import sys

import click

from GameserverLister.commands.options import common
from GameserverLister.common.logger import logger
from GameserverLister.common.types import Quake3Game
from GameserverLister.games.quake3 import QUAKE3_CONFIGS
from GameserverLister.listers import Quake3ServerLister


@click.command
@click.option(
    '-g',
    '--game',
    type=click.Choice(Quake3Game),
    required=True,
    help='Game to list servers for'
)
@click.option(
    '-p',
    '--principal',
    type=click.Choice([p for g in QUAKE3_CONFIGS for p in QUAKE3_CONFIGS[g]['servers'].keys()]),
    required=True,
    help='Principal server to query'
)
@common.expired_ttl
@common.list_dir
@common.recover
@common.add_links
@common.debug
def run(
        game: Quake3Game,
        principal: str,
        expired_ttl: int,
        recover: bool,
        add_links: bool,
        list_dir: str,
        debug: bool
):
    logging.basicConfig(level=logging.DEBUG if debug else logging.INFO, stream=sys.stdout,
                        format='%(asctime)s %(levelname)-8s %(message)s')
    logger.info(f'Listing servers for {game} via quake3/{principal}')

    # Set principal
    available_principals = list(QUAKE3_CONFIGS[game]['servers'].keys())
    if principal.lower() not in available_principals:
        # Given principal is invalid => use default principal
        principal = available_principals[0]

    lister = Quake3ServerLister(
        game,
        principal,
        expired_ttl,
        recover,
        add_links,
        list_dir
    )

    before = len(lister.servers)
    lister.update_server_list()
    removed, recovered = lister.remove_expired_servers()
    lister.write_to_file()

    logger.info(f'Server list updated ('
                f'total: {len(lister.servers)}, '
                f'added: {len(lister.servers) + removed - before}, '
                f'removed: {removed}, '
                f'recovered: {recovered})')



import logging
import sys
from random import randint
from typing import Tuple

import requests

from GameserverLister.common.helpers import bfbc2_server_validator, guid_from_ip_port
from GameserverLister.common.servers import BadCompany2Server
from GameserverLister.common.types import TheaterGame
from .common import FrostbiteServerLister


class BadCompany2ServerLister(FrostbiteServerLister):
    game: TheaterGame

    def __init__(self, expired_ttl: float, recover: bool, add_links: bool, list_dir: str, timeout: float):
        super().__init__(TheaterGame.BFBC2, BadCompany2Server, expired_ttl, recover, add_links, list_dir, timeout)
        self.server_validator = bfbc2_server_validator

    def update_server_list(self):
        request_ok = False
        attempt = 0
        max_attempts = 3
        servers = None
        while not request_ok and attempt < max_attempts:
            try:
                logging.info('Fetching server list from FESL/Theater API')
                resp = self.session.get('https://fesl.cetteup.com/servers/pc', timeout=self.request_timeout)

                if resp.ok:
                    servers = resp.json()
                    request_ok = True
                else:
                    attempt += 1
            except requests.exceptions.RequestException as e:
                logging.debug(e)
                logging.error(f'Failed to fetch servers from API, attempt {attempt + 1}/{max_attempts}')
                attempt += 1

        # Make sure any servers were found
        if servers is None:
            logging.error('Failed to retrieve server list, exiting')
            sys.exit(1)

        # Add servers from list
        found_servers = []
        for server in servers:
            found_server = BadCompany2Server(
                guid_from_ip_port(server['I'], server['P']),
                server['N'],
                int(server['LID']),
                int(server['GID']),
                server['I'],
                int(server['P'])
            )

            if self.add_links:
                found_server.add_links(self.build_server_links(
                    found_server.uid,
                    found_server.ip,
                    found_server.game_port
                ))

            found_servers.append(found_server)

        self.add_update_servers(found_servers)

    def check_if_server_still_exists(self, server: BadCompany2Server, checks_since_last_ok: int) -> Tuple[bool, bool, int]:
        check_ok = True
        found = False
        try:
            response = self.session.get(f'https://fesl.cetteup.com/servers/pc/{server.lid}/{server.gid}',
                                        timeout=self.request_timeout)

            if response.ok:
                found = True
            elif response.status_code == 404:
                found = False
            else:
                check_ok = False
            # Reset requests since last ok counter if server returned info/not found,
            # else increase counter and sleep
            if check_ok:
                checks_since_last_ok = 0
            else:
                checks_since_last_ok += 1
        except requests.RequestException as e:
            logging.debug(e)
            logging.error(f'Failed to fetch server {server.uid} for expiration check')
            check_ok = False

        return check_ok, found, checks_since_last_ok

    def build_port_to_try_list(self, game_port: int) -> list:
        """
        Most Bad Company 2 server seem to be hosted directly by members of the community, resulting in pretty random
        query ports as well as strange/incorrect server configurations. So, try a bunch of ports and validate found
        query ports using the connect property OR the server name
        Order of ports to try:
        1. default query port
        2. game port + default port offset (mirror gamedig behavior)
        3. game port (some servers use same port for game + query)
        4. game port + 100 (nitrado)
        5. game port + 10
        6. game port + 5 (several hosters)
        7. game port + 1
        8. game port + 29233 (i3D.net)
        9. game port + 29000
        10. game port + 29323
        11. random port around default query port
        12. random port around 48600
        13. random port between default game port and default query port
        14. random port between game port and game port + default offset
        """
        ports_to_try = [48888, game_port + 29321, game_port, game_port + 100, game_port + 10, game_port + 5,
                        game_port + 1, game_port + 29233, game_port + 29000, game_port + 29323, randint(48880, 48890),
                        randint(48601, 48605), randint(19567, 48888), randint(game_port, game_port + 29321)]

        return ports_to_try



import json
import logging
import os
import sys
import time
from datetime import datetime, timedelta
from random import choices
from typing import Type, List, Tuple, Optional, Union, Callable

import gevent
import requests
from gevent.pool import Pool

from GameserverLister.common.helpers import is_valid_port, find_query_port
from GameserverLister.common.servers import Server, ObjectJSONEncoder, FrostbiteServer
from GameserverLister.common.types import Game
from GameserverLister.common.weblinks import WebLink


class ServerLister:
    game: Game
    server_list_dir_path: str
    server_list_file_path: str
    expired_ttl: float
    recover: bool
    add_links: bool
    ensure_ascii: bool
    server_class: Type[Server]
    servers: List[Server]

    session: requests.Session
    request_timeout: float

    def __init__(
            self,
            game: Game,
            server_class: Type[Server],
            expired_ttl: float,
            recover: bool,
            add_links: bool,
            list_dir: str,
            request_timeout: float = 5.0
    ):
        self.game = game
        self.server_list_dir_path = os.path.realpath(list_dir)
        self.server_list_file_path = os.path.join(self.server_list_dir_path, f'{self.game}-servers.json')

        self.expired_ttl = expired_ttl
        self.recover = recover
        self.add_links = add_links

        self.ensure_ascii = True
        self.server_class = server_class
        self.servers = []

        # Init session
        self.session = requests.session()
        self.request_timeout = request_timeout

        # Create list dir if it does not exist
        if not os.path.isdir(self.server_list_dir_path):
            try:
                os.mkdir(self.server_list_dir_path)
            except IOError as e:
                logging.debug(e)
                logging.error(f'Failed to create missing server list directory at {self.server_list_dir_path}')
                sys.exit(1)

        # Init server list with servers from existing list or empty one
        if os.path.isfile(self.server_list_file_path):
            try:
                with open(self.server_list_file_path, 'r') as serverListFile:
                    logging.info('Loading existing server list')
                    self.servers = json.load(serverListFile, object_hook=self.server_class.load)
            except IOError as e:
                logging.debug(e)
                logging.error('Failed to read existing server list file')
                sys.exit(1)
            except json.decoder.JSONDecodeError as e:
                logging.debug(e)
                logging.error('Failed to parse existing server list file contents')
                sys.exit(1)

    def update_server_list(self):
        pass

    def add_update_servers(self, found_servers: List[Server]):
        # Add/update found servers to/in known servers
        logging.info(f'Updating server list with {len(found_servers)} found servers')
        for found_server in found_servers:
            known_server_uids = [s.uid for s in self.servers]
            # Update existing server entry or add new one
            if found_server.uid in known_server_uids:
                logging.debug(f'Found server {found_server.uid} already known, updating')
                index = known_server_uids.index(found_server.uid)
                self.servers[index].update(found_server)
                self.servers[index].trim(self.expired_ttl)
            else:
                logging.debug(f'Found server {found_server.uid} is new, adding')
                # Add new server entry
                self.servers.append(found_server)

    def remove_expired_servers(self) -> tuple:
        # Iterate over copy of server list and remove any expired servers from the (actual) server list
        logging.info(f'Checking expiration ttl for {len(self.servers)} servers')
        checks_since_last_ok = 0
        expired_servers_removed = 0
        expired_servers_recovered = 0
        for server in self.servers[:]:
            expired = datetime.now().astimezone() > server.last_seen_at + timedelta(hours=self.expired_ttl)
            if expired and self.recover:
                # Attempt to recover expired server by contacting/accessing it directly
                time.sleep(self.get_backoff_timeout(checks_since_last_ok))
                # Check if server can be accessed directly
                check_ok, found, checks_since_last_ok = self.check_if_server_still_exists(
                    server, checks_since_last_ok
                )

                # Remove server if request was sent successfully but server was not found
                if check_ok and not found:
                    logging.debug(f'Server {server.uid} has not been seen in '
                                  f'{self.expired_ttl} hours and could not be recovered, removing it')
                    self.servers.remove(server)
                    expired_servers_removed += 1
                elif check_ok and found:
                    logging.debug(f'Server {server.uid} did not appear in list but is still online, '
                                  f'updating last seen at')
                    index = self.servers.index(server)
                    self.servers[index].last_seen_at = datetime.now().astimezone()
                    self.servers[index].trim(self.expired_ttl)

                    expired_servers_recovered += 1
            elif expired:
                logging.debug(f'Server {server.uid} has not been seen in '
                              f'{self.expired_ttl} hours, removing it')
                self.servers.remove(server)
                expired_servers_removed += 1

        return expired_servers_removed, expired_servers_recovered

    def check_if_server_still_exists(self, server: Server, checks_since_last_ok: int) -> Tuple[bool, bool, int]:
        pass

    def build_server_links(
            self,
            uid: str,
            ip: Optional[str] = None,
            port: Optional[int] = None
    ) -> Union[List[WebLink], WebLink]:
        # Default to no links
        return []

    def get_backoff_timeout(self, checks_since_last_ok: int) -> int:
        # Default to no backoff
        return 0

    def write_to_file(self):
        logging.info(f'Writing {len(self.servers)} servers to output file')
        with open(self.server_list_file_path, 'w') as output_file:
            json.dump(self.servers, output_file, indent=2, ensure_ascii=self.ensure_ascii, cls=ObjectJSONEncoder)


class FrostbiteServerLister(ServerLister):
    servers: List[FrostbiteServer]
    server_validator: Callable

    def __init__(
            self,
            game: Game,
            server_class: Type[Server],
            expired_ttl: float,
            recover: bool,
            add_links: bool,
            list_dir: str,
            request_timeout: float = 5.0
    ):
        super().__init__(game, server_class, expired_ttl, recover, add_links, list_dir, request_timeout)

    def find_query_ports(self, gamedig_bin_path: str, gamedig_concurrency: int, expired_ttl: float):
        logging.info(f'Searching query port for {len(self.servers)} servers')

        search_stats = {
            'totalSearches': len(self.servers),
            'queryPortFound': 0,
            'queryPortReset': 0
        }
        pool = Pool(gamedig_concurrency)
        jobs = []
        for server in self.servers:
            ports_to_try = self.build_port_to_try_list(server.game_port)

            # Remove any invalid ports
            ports_to_try = [p for p in ports_to_try if is_valid_port(p)]

            # Get default port (first in list) and 5 random ports from selection
            ports_to_try = [ports_to_try[0], *choices(ports_to_try[1:], k=5)]

            jobs.append(
                pool.spawn(find_query_port, gamedig_bin_path, self.game, server, ports_to_try, self.server_validator)
            )
        # Wait for all jobs to complete
        gevent.joinall(jobs)
        for index, job in enumerate(jobs):
            server = self.servers[index]
            logging.debug(f'Checking query port search result for {server.uid}')
            if job.value != -1:
                logging.debug(f'Query port found ({job.value}), updating server')
                server.query_port = job.value
                server.last_queried_at = datetime.now().astimezone()
                search_stats['queryPortFound'] += 1
            elif server.query_port != -1 and \
                    (server.last_queried_at is None or
                     datetime.now().astimezone() > server.last_queried_at + timedelta(hours=expired_ttl)):
                logging.debug(f'Query port expired, resetting to -1 (was {server.query_port})')
                server.query_port = -1
                # TODO Reset last queried at here?
                search_stats['queryPortReset'] += 1
        logging.info(f'Query port search stats: {search_stats}')

    # Function has to be public to overrideable by derived classes
    def build_port_to_try_list(self, game_port: int) -> list:
        pass


class HttpServerLister(ServerLister):
    page_limit: int
    per_page: int
    sleep: float
    max_attempts: int

    def __init__(
            self,
            game: Game,
            server_class: Type[Server],
            page_limit: int,
            per_page: int,
            expired_ttl: float,
            recover: bool,
            add_links: bool,
            list_dir: str,
            sleep: float,
            max_attempts: int
    ):
        super().__init__(game, server_class, expired_ttl, recover, add_links, list_dir, request_timeout=10)
        self.page_limit = page_limit
        self.per_page = per_page
        self.sleep = sleep
        self.max_attempts = max_attempts

    def update_server_list(self):
        offset = 0
        """
        The Frostbite server browsers returns tons of duplicate servers (pagination is completely broken/non-existent).
        You basically just a [per_page] random servers every time. Thus, there is no way of telling when to stop.
        As a workaround, just stop after not retrieving a new/unique server for [args.page_limit] pages
        """
        pages_since_last_unique_server = 0
        attempt = 0
        """
        Since pagination of the server list is completely broken, just get the first "page" over and over again until
        no servers have been found in [args.page_limit] "pages".
        """
        found_servers = []
        logging.info('Starting server list retrieval')
        while pages_since_last_unique_server < self.page_limit and attempt < self.max_attempts:
            # Sleep when requesting anything but offset 0 (use increased sleep when retrying)
            if offset > 0:
                time.sleep(pow(self.sleep, attempt + 1))

            try:
                response = self.session.get(
                    self.get_server_list_url(self.per_page),
                    timeout=self.request_timeout
                )
            except requests.exceptions.RequestException as e:
                logging.debug(e)
                logging.error(f'Request failed, retrying {attempt + 1}/{self.max_attempts}')
                # Count try and start over
                attempt += 1
                continue

            if response.status_code == 200:
                # Reset tries
                attempt = 0
                # Parse response
                parsed = response.json()
                server_total_before = len(found_servers)
                # Add all servers in response (if they are new)
                found_servers = self.add_page_found_servers(found_servers, parsed)
                if len(found_servers) == server_total_before:
                    pages_since_last_unique_server += 1
                    logging.info(f'Got nothing but duplicates (page: {int(offset / self.per_page)},'
                                 f' pages since last unique: {pages_since_last_unique_server})')
                else:
                    logging.info(f'Got {len(found_servers) - server_total_before} new servers')
                    # Found new unique server, reset
                    pages_since_last_unique_server = 0
                offset += self.per_page
            else:
                logging.error(f'Server responded with {response.status_code}, '
                              f'retrying {attempt + 1}/{self.max_attempts}')
                attempt += 1

        self.add_update_servers(found_servers)

    def get_server_list_url(self, per_page: int) -> str:
        pass

    def add_page_found_servers(self, found_servers: List[Server], page_response_data: dict) -> List[Server]:
        pass

    def get_backoff_timeout(self, checks_since_last_ok: int) -> int:
        return 1 + pow(self.sleep, checks_since_last_ok % self.max_attempts)



import logging
import sys

import click

from GameserverLister.commands.options import common, http
from GameserverLister.common.logger import logger
from GameserverLister.common.types import GametoolsGame
from GameserverLister.listers import GametoolsServerLister


@click.command
@click.option(
    '-g',
    '--game',
    type=click.Choice(GametoolsGame),
    required=True,
    help='Game to list servers for'
)
@click.option(
    '--include-official',
    default=False,
    is_flag=True,
    help='Include DICE official servers in list (not recommended due to auto scaling official servers)'
)
@http.page_limit
@http.sleep
@http.max_attempts
@common.expired_ttl
@common.list_dir
@common.recover
@common.add_links
@common.debug
def run(
        game: GametoolsGame,
        page_limit: int,
        sleep: float,
        max_attempts: int,
        include_official: bool,
        expired_ttl: int,
        recover: bool,
        add_links: bool,
        list_dir: str,
        debug: bool
):
    logging.basicConfig(level=logging.DEBUG if debug else logging.INFO, stream=sys.stdout,
                        format='%(asctime)s %(levelname)-8s %(message)s')
    logger.info(f'Listing servers for {game} via gametools')

    lister = GametoolsServerLister(
        game,
        page_limit,
        expired_ttl,
        recover,
        add_links,
        list_dir,
        sleep,
        max_attempts,
        include_official
    )

    before = len(lister.servers)
    lister.update_server_list()

    removed, recovered = lister.remove_expired_servers()
    lister.write_to_file()

    logger.info(f'Server list updated ('
                f'total: {len(lister.servers)}, '
                f'added: {len(lister.servers) + removed - before}, '
                f'removed: {removed}, '
                f'recovered: {recovered})')



import logging
import sys
from typing import Tuple

import pyq3serverlist
import requests

from GameserverLister.common.helpers import is_valid_public_ip, is_valid_port, guid_from_ip_port
from GameserverLister.common.servers import ClassicServer, ViaStatus
from GameserverLister.common.types import MedalOfHonorGame
from .common import ServerLister


class MedalOfHonorServerLister(ServerLister):
    """
    The defacto standard principal server for Medal of Honor games uses a query GameSpy protocol implementation that
    does not work with gslist. However, they also provide server lists as text files, which unfortunately only contain
    the game port. So, instead of extending the GameSpy server lister and trying to find the query port, we use the
    text files to find servers. Since the Medal of Honor games support Quake3-like queries on the game port, we can use
    that to query servers with the information we have.
    """
    game: MedalOfHonorGame

    def __init__(self, game: MedalOfHonorGame, expired_ttl: float, recover: bool, add_links: bool, list_dir: str):
        super().__init__(game, ClassicServer, expired_ttl, recover, add_links, list_dir)

    def update_server_list(self):
        request_ok = False
        attempt = 0
        max_attempts = 3
        raw_server_list = None
        while not request_ok and attempt < max_attempts:
            try:
                logging.info('Fetching server list from mohaaservers.tk')
                resp = self.session.get(self.get_server_list_url(), timeout=self.request_timeout)

                if resp.ok:
                    raw_server_list = resp.text
                    request_ok = True
                else:
                    attempt += 1
            except requests.exceptions.RequestException as e:
                logging.debug(e)
                logging.error(f'Failed to fetch servers from mohaaservers.tk, attempt {attempt + 1}/{max_attempts}')
                attempt += 1

        # Make sure any servers were found
        if raw_server_list is None:
            logging.error('Failed to retrieve server list, exiting')
            sys.exit(1)

        # Parse list
        found_servers = []
        for line in raw_server_list.strip(' \n').split('\n'):
            # Silently skip any completely empty lines
            if len(line) == 0:
                continue

            elems = line.strip().split(':')

            if len(elems) != 2:
                logging.warning(f'Principal returned malformed server list entry ({line}), skipping it')
                continue

            ip, query_port = elems
            if not is_valid_public_ip(ip) or not is_valid_port(int(query_port)):
                logging.warning(f'Principal returned invalid server entry ({ip}:{query_port}), skipping it')
                continue

            via = ViaStatus('mohaaservers.tk')
            found_server = ClassicServer(
                guid_from_ip_port(ip, query_port),
                ip,
                int(query_port),
                via
            )

            if self.add_links:
                # query port = game port for MOH games, so we can use that here to build links
                found_server.add_links(self.build_server_links(
                    found_server.uid,
                    found_server.ip,
                    found_server.query_port
                ))

            found_servers.append(found_server)

        self.add_update_servers(found_servers)

    def get_server_list_url(self) -> str:
        # mohaaservers uses a "key" without the "moh" prefix, e.g. "servers_aa" for "mohaa"
        return f'https://mohaaservers.tk/servlist/servers_{self.game[3:]}.txt'

    def check_if_server_still_exists(self, server: ClassicServer, checks_since_last_ok: int) -> Tuple[bool, bool, int]:
        # Since we query the server directly, there is no way of handling HTTP server errors differently then
        # actually failed checks, so even if the query fails, we have to treat it as "check ok"
        check_ok = True
        found = False
        try:
            pyq3serverlist.MedalOfHonorServer(server.ip, server.query_port).get_status()

            # get_status will raise an exception if the server cannot be contacted,
            # so this will only be reached if the query succeeds
            found = True
        except pyq3serverlist.PyQ3SLError as e:
            logging.debug(e)
            logging.debug(f'Failed to query server {server.uid} for expiration check')

        return check_ok, found, checks_since_last_ok



import click

expired_ttl = click.option(
    '-e',
    '--expired-ttl',
    type=float,
    default=12.0,
    help='How long to keep a server in list after it was last seen (in hours)'
)
recover = click.option(
    '--no-recover',
    'recover',
    default=True,
    is_flag=True,
    help='Remove servers that were not returned by the source after they expired, '
         'do not attempt to contact/access server directly to check if they are still online'
)
add_links = click.option(
    '--add-links',
    default=False,
    is_flag=True,
    help='Enrich server list entries with links to websites showing more details about the server'
)
list_dir = click.option(
    '-d',
    '--list-dir',
    type=str,
    default='lists',
    help='Path to directory in which servers lists will be stored'
)
debug = click.option(
    '--debug',
    default=False,
    is_flag=True,
    help='Log lots of debugging information',
)



import logging
from datetime import datetime
from typing import List, Tuple, Optional, Union

import requests

from GameserverLister.common.servers import GametoolsServer
from GameserverLister.common.types import GametoolsGame
from GameserverLister.common.weblinks import WebLink, WEB_LINK_TEMPLATES
from GameserverLister.games.gametools import GAMETOOLS_BASE_URI
from .common import HttpServerLister


class GametoolsServerLister(HttpServerLister):
    game: GametoolsGame
    include_official: bool

    def __init__(
            self,
            game: GametoolsGame,
            page_limit: int,
            expired_ttl: float,
            recover: bool,
            add_links: bool,
            list_dir: str,
            sleep: float,
            max_attempts: int,
            include_official: bool
    ):
        super().__init__(game, GametoolsServer, page_limit, 100, expired_ttl, recover, add_links, list_dir, sleep,
                         max_attempts)
        # Allow non-ascii characters in server list (mostly used by server names for Asia servers)
        self.ensure_ascii = False
        self.include_official = include_official

    def get_server_list_url(self, per_page: int) -> str:
        return f'{GAMETOOLS_BASE_URI}/{self.game}/servers/?name=&limit={per_page}' \
               f'&nocache={datetime.now().timestamp()}'

    def add_page_found_servers(self, found_servers: List[GametoolsServer], page_response_data: dict) -> List[GametoolsServer]:
        for server in page_response_data['servers']:
            found_server = GametoolsServer(
                server['gameId'],
                server['prefix'],
            )

            if self.add_links:
                found_server.add_links(self.build_server_links(found_server.uid))

            # Add/update servers (ignoring official servers unless include_official is set)
            server_game_ids = [s.uid for s in found_servers]
            if found_server.uid not in server_game_ids and \
                    (not server['official'] or self.include_official):
                logging.debug(f'Got new server {found_server.uid}, adding it')
                found_servers.append(found_server)
            elif not server['official'] or self.include_official:
                logging.debug(f'Got duplicate server {found_server.uid}, updating last seen at')
                index = server_game_ids.index(found_server.uid)
                found_servers[index].last_seen_at = datetime.now().astimezone()
            else:
                logging.debug(f'Got official server {found_server.uid}, ignoring it')

        return found_servers

    def check_if_server_still_exists(self, server: GametoolsServer, checks_since_last_ok: int) -> Tuple[bool, bool, int]:
        check_ok = True
        found = False
        try:
            response = self.session.get(f'{GAMETOOLS_BASE_URI}/{self.game}/detailedserver/'
                                        f'?gameid={server.uid}', timeout=self.request_timeout)
            if response.status_code == 200:
                # Server was found on gametools => make sure it still not official (or include_official is set)
                parsed = response.json()
                found = not parsed['official'] or self.include_official
            elif response.status_code == 404:
                # gametools responded with, server was not found
                found = False
            else:
                # gametools responded with some other status code (504 gateway timeout for example)
                check_ok = False
            # Reset requests since last ok counter if server returned info/not found,
            # else increase counter and sleep
            if check_ok:
                checks_since_last_ok = 0
            else:
                checks_since_last_ok += 1
        except requests.exceptions.RequestException as e:
            logging.debug(e)
            logging.error(f'Failed to fetch server {server.uid} for expiration check')
            check_ok = False

        return check_ok, found, checks_since_last_ok

    def build_server_links(
            self,
            uid: str,
            ip: Optional[str] = None,
            port: Optional[int] = None
    ) -> Union[List[WebLink], WebLink]:
        links = [WEB_LINK_TEMPLATES['gametools'].render(self.game, uid)]
        # BF1 servers are also listed on battlefieldtracker.com
        if self.game is GametoolsGame.BF1:
            links.append(WEB_LINK_TEMPLATES['battlefieldtracker'].render(self.game, uid))

        return links



from typing import Dict

from GameserverLister.common.types import Unreal2Game

UNREAL2_CONFIGS: Dict[Unreal2Game, dict] = {
    Unreal2Game.UT2003: {
        'servers': {
            'openspy.net': {
                'hostname': 'utmaster.openspy.net',
                'port': 28902
            }
        }
    },
    Unreal2Game.UT2004: {
        'servers': {
            'openspy.net': {
                'hostname': 'utmaster.openspy.net',
                'port': 28902
            },
            '333networks.com': {
                'hostname': 'ut2004master.333networks.com',
                'port': 28902
            },
            'errorist.eu': {
                'hostname': 'ut2004master.errorist.eu',
                'port': 28902
            }
        }
    }
}



from typing import Dict

from GameserverLister.common.types import GamespyPrincipal, GamespyPrincipalConfig, GamespyGame, GamespyGameConfig

GAMESPY_PRINCIPAL_CONFIGS: Dict[GamespyPrincipal, GamespyPrincipalConfig] = {
    GamespyPrincipal.TripleThreeNetworks_com_1: GamespyPrincipalConfig(
        hostname='master.333networks.com'
    ),
    GamespyPrincipal.TripleThreeNetworks_com_2: GamespyPrincipalConfig(
        hostname='rhea.333networks.com'
    ),
    GamespyPrincipal.BF1942_org: GamespyPrincipalConfig(
        hostname='master.bf1942.org'
    ),
    GamespyPrincipal.BF2Hub_com: GamespyPrincipalConfig(
        hostname='{0}.master.bf2hub.com',
        port_offset=1
    ),
    GamespyPrincipal.Crymp_net: GamespyPrincipalConfig(
        hostname='master.crymp.net'
    ),
    GamespyPrincipal.Errorist_eu: GamespyPrincipalConfig(
        hostname='master.errorist.eu'
    ),
    GamespyPrincipal.FH2_dev: GamespyPrincipalConfig(
        hostname='ms.fh2.dev'
    ),
    GamespyPrincipal.Jedi95_us: GamespyPrincipalConfig(
        hostname='master.g.jedi95.us'
    ),
    GamespyPrincipal.Newbiesplayground_net: GamespyPrincipalConfig(
        hostname='master.newbiesplayground.net'
    ),
    GamespyPrincipal.NightfirePC_com: GamespyPrincipalConfig(
        hostname='master.nightfirepc.com'  # This (currently) just points at openspy
    ),
    GamespyPrincipal.NovGames_ru: GamespyPrincipalConfig(
        hostname='2142.novgames.ru'
    ),
    GamespyPrincipal.OldUnreal_com_1: GamespyPrincipalConfig(
        hostname='master.oldunreal.com'
    ),
    GamespyPrincipal.OldUnreal_com_2: GamespyPrincipalConfig(
        hostname='master2.oldunreal.com'
    ),
    GamespyPrincipal.OpenSpy_net: GamespyPrincipalConfig(
        hostname='{0}.master.openspy.net'
    ),
    GamespyPrincipal.PhoenixNetwork_net: GamespyPrincipalConfig(
        hostname='master.phoenixnetwork.net'
    ),
    GamespyPrincipal.Play2142_ru: GamespyPrincipalConfig(
        hostname='{0}.ms.play2142.ru'
    ),
    GamespyPrincipal.PlayBF2_ru: GamespyPrincipalConfig(
        hostname='{0}.ms.playbf2.ru'
    ),
    GamespyPrincipal.Qtracker_com: GamespyPrincipalConfig(
        hostname='master2.qtracker.com'
    ),
    GamespyPrincipal.SWAT4Stats_com: GamespyPrincipalConfig(
        hostname='master.swat4stats.com'
    ),
    GamespyPrincipal.Vietcong_tk: GamespyPrincipalConfig(
        hostname='51.68.46.73' # From http://vcfiles.free.fr/r.php?r=files/patch_serverlist.bat, no hostname available
    ),
    GamespyPrincipal.Vietcong1_eu: GamespyPrincipalConfig(
        hostname='vietcong1.eu'
    )
}
GAMESPY_GAME_CONFIGS: Dict[GamespyGame, GamespyGameConfig] = {
    GamespyGame.BF1942: GamespyGameConfig(
        game_name='bfield1942',
        game_key='HpWx9z',
        enc_type=2,
        query_type=0,
        port=28900,
        principals=[
            GamespyPrincipal.BF1942_org,
            GamespyPrincipal.OpenSpy_net,
            GamespyPrincipal.Qtracker_com
        ]
    ),
    GamespyGame.BFVietnam: GamespyGameConfig(
        game_name='bfvietnam',
        game_key='h2P9dJ',
        enc_type=2,
        query_type=0,
        port=28900,
        principals=[
            GamespyPrincipal.OpenSpy_net,
            GamespyPrincipal.Qtracker_com
        ]
    ),
    GamespyGame.BF2: GamespyGameConfig(
        game_name='battlefield2',
        game_key='hW6m9a',
        enc_type=-1,
        query_type=8,
        port=28910,
        principals=[
            GamespyPrincipal.BF2Hub_com,
            GamespyPrincipal.OpenSpy_net,
            GamespyPrincipal.PhoenixNetwork_net,
            GamespyPrincipal.PlayBF2_ru
        ],
        link_template_refs={
            '_any': ['bf2.tv'],
            GamespyPrincipal.BF2Hub_com: ['bf2hub']
        }
    ),
    GamespyGame.FH2: GamespyGameConfig(
        game_name='battlefield2',
        game_key='hW6m9a',
        enc_type=-1,
        query_type=8,
        port=28910,
        principals=[
            GamespyPrincipal.FH2_dev
        ],
        info_query='\\hostname'
    ),
    GamespyGame.BF2142: GamespyGameConfig(
        game_name='stella',
        game_key='M8o1Qw',
        enc_type=-1,
        query_type=8,
        port=28910,
        principals=[
            GamespyPrincipal.NovGames_ru,
            GamespyPrincipal.OpenSpy_net,
            GamespyPrincipal.Play2142_ru
        ]
    ),
    GamespyGame.Crysis: GamespyGameConfig(
        game_name='crysis',
        game_key='ZvZDcL',
        enc_type=-1,
        query_type=8,
        port=28910,
        principals=[
            GamespyPrincipal.Crymp_net
        ]
    ),
    GamespyGame.CrysisWars: GamespyGameConfig(
        game_name='crysiswars',
        game_key='zKbZiM',
        enc_type=-1,
        query_type=8,
        port=28910,
        principals=[
            GamespyPrincipal.Jedi95_us
        ]
    ),
    GamespyGame.DeusEx: GamespyGameConfig(
        game_name='deusex',
        game_key='Av3M99',
        enc_type=0,
        query_type=0,
        port=28900,
        principals=[
            GamespyPrincipal.TripleThreeNetworks_com_1,
            GamespyPrincipal.Errorist_eu,
            GamespyPrincipal.Newbiesplayground_net,
            GamespyPrincipal.OldUnreal_com_1
        ]
    ),
    GamespyGame.DukeNukemForever: GamespyGameConfig(
        game_name='dnf',
        game_key='      ',
        enc_type=0,
        query_type=0,
        port=28900,
        principals=[
            GamespyPrincipal.TripleThreeNetworks_com_1
        ]
    ),
    GamespyGame.JBNightfire: GamespyGameConfig(
        game_name='jbnightfire',
        game_key='S9j3L2',
        enc_type=-1,
        query_type=0,
        port=28910,
        principals=[
            GamespyPrincipal.OpenSpy_net,
            GamespyPrincipal.NightfirePC_com
        ]
    ),
    GamespyGame.Paraworld: GamespyGameConfig(
        game_name='paraworld',
        game_key='EUZpQF',
        enc_type=-1,
        query_type=8,
        port=28910,
        principals=[
            GamespyPrincipal.OpenSpy_net
        ]
    ),
    GamespyGame.Postal2: GamespyGameConfig(
        game_name='postal2',
        game_key='yw3R9c',
        enc_type=0,
        query_type=0,
        port=28900,
        principals=[
            GamespyPrincipal.TripleThreeNetworks_com_1
        ]
    ),
    GamespyGame.Rune: GamespyGameConfig(
        game_name='rune',
        game_key='BnA4a3',
        enc_type=0,
        query_type=0,
        port=28900,
        principals=[
            GamespyPrincipal.TripleThreeNetworks_com_1,
            GamespyPrincipal.Errorist_eu,
            GamespyPrincipal.Newbiesplayground_net,
            GamespyPrincipal.OldUnreal_com_1
        ]
    ),
    GamespyGame.SeriousSam: GamespyGameConfig(
        game_name='serioussam',
        game_key='AKbna4',
        enc_type=0,
        query_type=0,
        port=28900,
        principals=[
            GamespyPrincipal.TripleThreeNetworks_com_1,
            GamespyPrincipal.Errorist_eu,
            GamespyPrincipal.Newbiesplayground_net,
            GamespyPrincipal.OldUnreal_com_1
        ]
    ),
    GamespyGame.SeriousSamSE: GamespyGameConfig(
        game_name='serioussamse',
        game_key='AKbna4',
        enc_type=0,
        query_type=0,
        port=28900,
        principals=[
            GamespyPrincipal.TripleThreeNetworks_com_1,
            GamespyPrincipal.Errorist_eu,
            GamespyPrincipal.Newbiesplayground_net,
            GamespyPrincipal.OldUnreal_com_1
        ]
    ),
    GamespyGame.SWAT4: GamespyGameConfig(
        game_name='swat4',
        game_key='tG3j8c',
        enc_type=-1,
        query_type=0,
        port=28910,
        principals=[
            GamespyPrincipal.SWAT4Stats_com
        ],
        info_query='\\hostname',
        link_template_refs={
            GamespyPrincipal.SWAT4Stats_com: ['swat4stats.com']
        }
    ),
    GamespyGame.Unreal: GamespyGameConfig(
        game_name='unreal',
        game_key='DAncRK',
        enc_type=0,
        query_type=0,
        port=28900,
        principals=[
            GamespyPrincipal.TripleThreeNetworks_com_1,
            GamespyPrincipal.OldUnreal_com_1,
            GamespyPrincipal.Errorist_eu,
            GamespyPrincipal.OpenSpy_net,
            GamespyPrincipal.Qtracker_com
        ]
    ),
    GamespyGame.UT: GamespyGameConfig(
        game_name='ut',
        game_key='Z5Nfb0',
        enc_type=0,
        query_type=0,
        port=28900,
        principals=[
            GamespyPrincipal.TripleThreeNetworks_com_1,
            GamespyPrincipal.OldUnreal_com_1,
            GamespyPrincipal.Errorist_eu,
            GamespyPrincipal.OpenSpy_net,
            GamespyPrincipal.Qtracker_com
        ]
    ),
    GamespyGame.UT3: GamespyGameConfig(
        game_name='ut3pc',
        game_key='nT2Mtz',
        enc_type=-1,
        query_type=11,
        port=28910,
        principals=[
            GamespyPrincipal.OpenSpy_net
        ]
    ),
    GamespyGame.Vietcong: GamespyGameConfig(
        game_name='vietcong',
        game_key='bq98mE',
        enc_type=2,
        query_type=0,
        port=28900,
        principals=[
            GamespyPrincipal.Vietcong_tk,
            GamespyPrincipal.Vietcong1_eu,
            GamespyPrincipal.Qtracker_com
        ]
    ),
    GamespyGame.Vietcong2: GamespyGameConfig(
        game_name='vietcong2',
        game_key='zX2pq6',
        enc_type=-1,
        query_type=8,
        port=28910,
        principals=[
            GamespyPrincipal.OpenSpy_net
        ]
    ),
    GamespyGame.WheelOfTime: GamespyGameConfig(
        game_name='wot',
        game_key='RSSSpA',
        enc_type=0,
        query_type=0,
        port=28900,
        principals=[
            GamespyPrincipal.TripleThreeNetworks_com_1,
            GamespyPrincipal.Errorist_eu,
            GamespyPrincipal.Newbiesplayground_net,
            GamespyPrincipal.OldUnreal_com_1
        ]
    ),
}



import logging
import socket
from typing import List, Tuple, Optional, Union

import pyq3serverlist

from GameserverLister.common.helpers import is_valid_public_ip, is_valid_port, guid_from_ip_port
from GameserverLister.common.servers import ClassicServer, ViaStatus
from GameserverLister.common.types import Quake3Game
from GameserverLister.common.weblinks import WebLink, WEB_LINK_TEMPLATES
from GameserverLister.games.quake3 import QUAKE3_CONFIGS
from .common import ServerLister


class Quake3ServerLister(ServerLister):
    game: Quake3Game
    principal: str
    protocols: List[int]
    network_protocol: int
    game_name: str
    keywords: str
    server_entry_prefix: bytes

    def __init__(self, game: Quake3Game, principal: str, expired_ttl: float, recover: bool, add_links: bool, list_dir: str):
        super().__init__(game, ClassicServer, expired_ttl, recover, add_links, list_dir)
        # Merge default config with given principal config
        default_config = {
            'keywords': 'full empty',
            'game_name': '',
            'network_protocol': socket.SOCK_DGRAM,
            'server_entry_prefix': b''
        }
        principal_config = {key: value for (key, value) in QUAKE3_CONFIGS[self.game].items()
                            if key in default_config.keys()}
        self.principal = principal
        self.keywords, self.game_name, \
            self.network_protocol, self.server_entry_prefix = {**default_config, **principal_config}.values()
        self.protocols = QUAKE3_CONFIGS[self.game]['protocols']

    def update_server_list(self):
        # Use same connection to principal for all queries
        hostname, port = QUAKE3_CONFIGS[self.game]['servers'][self.principal].values()
        principal = pyq3serverlist.PrincipalServer(hostname, port, self.network_protocol)

        # Fetch servers for all protocols
        found_servers = []
        for protocol in self.protocols:
            raw_servers = self.get_servers(principal, protocol)
            for raw_server in raw_servers:
                if not is_valid_public_ip(raw_server.ip) or not is_valid_port(raw_server.port):
                    logging.warning(
                        f'Principal returned invalid server entry '
                        f'({raw_server.ip}:{raw_server.port}), skipping it'
                    )
                    continue

                via = ViaStatus(self.principal)
                found_server = ClassicServer(
                    guid_from_ip_port(raw_server.ip, str(raw_server.port)),
                    raw_server.ip,
                    raw_server.port,
                    via
                )

                if self.add_links:
                    found_server.add_links(self.build_server_links(
                        found_server.uid,
                        found_server.ip,
                        found_server.query_port
                    ))

                found_servers.append(found_server)

        self.add_update_servers(found_servers)

    def get_servers(self, principal: pyq3serverlist.PrincipalServer, protocol: int) -> List[pyq3serverlist.Server]:
        query_ok = False
        attempt = 0
        max_attempts = 3
        servers = []
        while not query_ok and attempt < max_attempts:
            try:
                servers = principal.get_servers(protocol, self.game_name, self.keywords, self.server_entry_prefix)
                query_ok = True
            except pyq3serverlist.PyQ3SLTimeoutError:
                logging.error(f'Principal server query timed out using protocol {protocol}, '
                              f'attempt {attempt + 1}/{max_attempts}')
                attempt += 1
            except pyq3serverlist.PyQ3SLError as e:
                logging.debug(e)
                logging.error(f'Failed to query principal server using protocol {protocol}, '
                              f'attempt {attempt + 1}/{max_attempts}')
                attempt += 1

        return servers

    def check_if_server_still_exists(self, server: ClassicServer, checks_since_last_ok: int) -> Tuple[bool, bool, int]:
        # Since we query the server directly, there is no way of handling HTTP server errors differently then
        # actually failed checks, so even if the query fails, we have to treat it as "check ok"
        check_ok = True
        found = False
        try:
            pyq3serverlist.Server(server.ip, server.query_port).get_status()

            # get_status will raise an exception if the server cannot be contacted,
            # so this will only be reached if the query succeeds
            found = True
        except pyq3serverlist.PyQ3SLError as e:
            logging.debug(e)
            logging.debug(f'Failed to query server {server.uid} for expiration check')

        return check_ok, found, checks_since_last_ok

    def build_server_links(
            self,
            uid: str,
            ip: Optional[str] = None,
            port: Optional[int] = None
    ) -> Union[List[WebLink], WebLink]:
        template_refs = QUAKE3_CONFIGS[self.game].get('linkTemplateRefs', {})
        # Add principal-scoped links first, then add game-scoped links
        templates = [
            *[WEB_LINK_TEMPLATES.get(ref) for ref in template_refs.get(self.principal, [])],
            *[WEB_LINK_TEMPLATES.get(ref) for ref in template_refs.get('_any', [])]
        ]

        links = []
        for template in templates:
            links.append(template.render(self.game, uid, ip=ip, port=port))

        return links



import logging
from typing import List, Tuple, Optional, Union

import pyut2serverlist

from GameserverLister.common.helpers import is_valid_public_ip, is_valid_port, guid_from_ip_port
from GameserverLister.common.servers import ClassicServer, ViaStatus
from GameserverLister.common.types import Unreal2Game
from GameserverLister.common.weblinks import WebLink, WEB_LINK_TEMPLATES
from GameserverLister.games.unreal2 import UNREAL2_CONFIGS
from .common import ServerLister


class Unreal2ServerLister(ServerLister):
    game: Unreal2Game
    principal: str
    cd_key: str

    principal_timeout: float

    def __init__(
            self,
            game: Unreal2Game,
            principal: str,
            cd_key: str,
            principal_timeout: float,
            expired_ttl: float,
            recover: bool,
            add_links: bool,
            list_dir: str
    ):
        super().__init__(game, ClassicServer, expired_ttl, recover, add_links, list_dir)
        self.principal = principal
        self.cd_key = cd_key
        self.principal_timeout = principal_timeout

    def update_server_list(self):
        hostname, port = UNREAL2_CONFIGS[self.game]['servers'][self.principal].values()
        principal = pyut2serverlist.PrincipalServer(
            hostname,
            port,
            pyut2serverlist.Game(self.game),
            self.cd_key,
            timeout=self.principal_timeout
        )

        found_servers = []
        raw_servers = self.get_servers(principal)
        for raw_server in raw_servers:
            if not is_valid_public_ip(raw_server.ip) or not is_valid_port(raw_server.query_port):
                logging.warning(
                    f'Principal returned invalid server entry '
                    f'({raw_server.ip}:{raw_server.query_port}), skipping it'
                )
                continue

            via = ViaStatus(self.principal)
            found_server = ClassicServer(
                guid_from_ip_port(raw_server.ip, str(raw_server.query_port)),
                raw_server.ip,
                raw_server.query_port,
                via
            )

            if self.add_links:
                found_server.add_links(self.build_server_links(
                    found_server.uid,
                    found_server.ip,
                    raw_server.game_port
                ))

            found_servers.append(found_server)

        self.add_update_servers(found_servers)

    @staticmethod
    def get_servers(principal: pyut2serverlist.PrincipalServer) -> List[pyut2serverlist.Server]:
        query_ok = False
        attempt = 0
        max_attempts = 3
        servers = []
        while not query_ok and attempt < max_attempts:
            try:
                servers = principal.get_servers()
                query_ok = True
            except pyut2serverlist.TimeoutError:
                logging.error(f'Principal server query timed out, attempt {attempt + 1}/{max_attempts}')
                attempt += 1
            except pyut2serverlist.Error as e:
                logging.debug(e)
                logging.error(f'Failed to query principal server, attempt {attempt + 1}/{max_attempts}')
                attempt += 1

        return servers

    def check_if_server_still_exists(self, server: ClassicServer, checks_since_last_ok: int) -> Tuple[bool, bool, int]:
        # Since we query the server directly, there is no way of handling HTTP server errors differently then
        # actually failed checks, so even if the query fails, we have to treat it as "check ok"
        check_ok = True
        found = False
        try:
            pyut2serverlist.Server(server.ip, server.query_port).get_info()

            # get_info will raise an exception if the server cannot be contacted,
            # so this will only be reached if the query succeeds
            found = True
        except pyut2serverlist.Error as e:
            logging.debug(e)
            logging.debug(f'Failed to query server {server.uid} for expiration check')

        return check_ok, found, checks_since_last_ok

    def build_server_links(
            self,
            uid: str,
            ip: Optional[str] = None,
            port: Optional[int] = None
    ) -> Union[List[WebLink], WebLink]:
        template_refs = UNREAL2_CONFIGS[self.game].get('linkTemplateRefs', {})
        # Add principal-scoped links first, then add game-scoped links
        templates = [
            *[WEB_LINK_TEMPLATES.get(ref) for ref in template_refs.get(self.principal, [])],
            *[WEB_LINK_TEMPLATES.get(ref) for ref in template_refs.get('_any', [])]
        ]

        links = []
        for template in templates:
            links.append(template.render(self.game, uid, ip=ip, port=port))

        return links
